__all__ = ['ShapDependenceTab']

import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_table

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from .dashboard_methods import *

class ShapDependenceTab:
    def __init__(self, explainer, standalone=False, tab_id="shap_dependence", title='Shap Dependence',
                 n_features=10, **kwargs):
        self.explainer = explainer
        self.standalone = standalone
        self.tab_id = tab_id
        self.title = title

        self.n_features = n_features
        self.kwargs = kwargs

        if self.standalone:
            self.label_selector = TitleAndLabelSelector(explainer, title=title)
             
    def layout(self):
        return dbc.Container([
            self.label_selector.layout() if self.standalone else None,
            # need to add dummy to make callbacks on tab change work:
            html.Div(id='tabs') if self.standalone else None, 
            shap_dependence_layout(self.explainer, n_features=self.n_features)
    
        ], fluid=True)
    
    def register_callbacks(self, app):
        if self.standalone:
            self.label_selector.register_callbacks(app)
        shap_dependence_callbacks(self.explainer, app)


def shap_dependence_layout(explainer, n_features=None, cats=True, **kwargs):

    cats_display = 'none' if explainer.cats is None else 'inline-block'
    return dbc.Container([
    dbc.Row([
        dbc.Col([
            html.H3('Shap Summary'),
            dbc.Row([
                dbc.Col([
                    dbc.Label("Depth:"),
                    dcc.Dropdown(id='dependence-scatter-depth',
                        options=[{'label': str(i+1), 'value': i+1} 
                                        for i in range(len(explainer.columns_ranked_by_shap(cats)))],
                        value=min(n_features, len(explainer.columns_ranked_by_shap(cats))))],
                    md=3), 
                dbc.Col([
                    dbc.FormGroup(
                        [
                            dbc.Label("Summary Type"),
                            dbc.RadioItems(
                                options=[
                                    {"label": "Aggregate", "value": "aggregate"},
                                    {"label": "Detailed", "value": "detailed"},
                                ],
                                value="aggregate",
                                id="dependence-summary-type",
                                inline=True,
                            ),
                        ]
                    )
                ], md=3),
                dbc.Col([
                    dbc.Label("Grouping:"),
                    dbc.FormGroup(
                    [
                        dbc.RadioButton(
                            id='dependence-group-categoricals', 
                            className="form-check-input",
                            checked=True),
                        dbc.Label("Group Cats",
                                html_for='dependence-group-categoricals',
                                className="form-check-label"),
                    ], check=True)],
                    md=3),
                ], form=True, justify="between"),

            dbc.Label('(Click on a dot to display dependece graph)'),
            dcc.Loading(id="loading-dependence-shap-summary", 
                    children=[dcc.Graph(id='dependence-shap-summary-graph')])
        ], md=6),
        dbc.Col([
            html.H3('Shap Dependence Plot'),
            dbc.Row([
                dbc.Col([
                    html.Label('Plot dependence for column:'),
                    dcc.Dropdown(id='dependence-col', 
                        options=[{'label': col, 'value':col} 
                                    for col in explainer.columns_ranked_by_shap(cats)],
                        value=explainer.columns_ranked_by_shap(cats)[0])],
                    md=5), 
                dbc.Col([
                     html.Label('Color observation by column:'),
                    dcc.Dropdown(id='dependence-color-col', 
                        options=[{'label': col, 'value':col} 
                                    for col in explainer.columns_ranked_by_shap(cats)],
                        value=explainer.columns_ranked_by_shap(cats)[1])],
                    md=5), 
                dbc.Col([
                    html.Label('Highlight:'),
                    dbc.Input(id='dependence-highlight-index', 
                            placeholder="Highlight index...",
                            debounce=True)]
                    , md=2) 
                ], form=True),
            
            dcc.Loading(id="loading-dependence-graph", 
                         children=[dcc.Graph(id='dependence-graph')]),
        ], md=6),
        ]),
    ],  fluid=True)


def shap_dependence_callbacks(explainer, app, **kwargs):
    
    @app.callback(
        [Output('dependence-shap-summary-graph', 'figure'),
         Output('dependence-col', 'options'),
         Output('dependence-scatter-depth', 'options')],
        [Input('dependence-summary-type', 'value'),
         Input('dependence-group-categoricals', 'checked'),
         Input('dependence-scatter-depth', 'value'),
         Input('label-store', 'data')],
        [State('tabs', 'value')])
    def update_dependence_shap_scatter_graph(summary_type, cats, depth, pos_label, tab):
        explainer.pos_label = pos_label #needed in case of multiple workers
        ctx = dash.callback_context
        if ctx.triggered:
            if summary_type=='aggregate':
                plot = explainer.plot_importances(
                        kind='shap', topx=depth, cats=cats)
            elif summary_type=='detailed':
                plot = explainer.plot_shap_summary(topx=depth, cats=cats)

            trigger = ctx.triggered[0]['prop_id'].split('.')[0]

            if trigger=='dependence-group-categoricals':
                # if change to group cats, adjust columns and depth
                if cats:
                    col_options = [{'label':col, 'value':col} 
                                for col in explainer.columns_cats]
                else:
                    col_options = [{'label':col, 'value':col} 
                                for col in explainer.columns]

                depth_options = [{'label': str(i+1), 'value':i+1} 
                            for i in range(len(col_options))]
                return (plot, col_options, depth_options)
                
            else:
                return (plot, dash.no_update, dash.no_update)
                
        raise PreventUpdate

    @app.callback(
        [Output('dependence-highlight-index', 'value'),
         Output('dependence-col', 'value')],
        [Input('dependence-shap-summary-graph', 'clickData')],
        [State('dependence-group-categoricals', 'checked')])
    def display_scatter_click_data(clickdata, cats):
        if clickdata is not None and clickdata['points'][0] is not None:
            if isinstance(clickdata['points'][0]['y'], float): # detailed
                # if detailed, clickdata returns scatter marker location -> type==float
                idx = clickdata['points'][0]['pointIndex']
                col = clickdata['points'][0]['text'].split('=')[0]                             
                return (idx, col)
            elif isinstance(clickdata['points'][0]['y'], str): # aggregate
                # in aggregate clickdata returns col name -> type==str
                col = clickdata['points'][0]['y'].split(' ')[1]
                return (dash.no_update, col) 
        raise PreventUpdate

    @app.callback(
        [Output('dependence-color-col', 'options'),
         Output('dependence-color-col', 'value')],
        [Input('dependence-col', 'value')],
        [State('dependence-group-categoricals', 'checked')])
    def set_color_col_dropdown(col, cats):
        sorted_interact_cols = explainer.shap_top_interactions(col, cats=cats)
        options = [{'label': col, 'value':col} 
                                    for col in sorted_interact_cols]
        value = sorted_interact_cols[1]                                
        return (options, value)


    @app.callback(
        Output('dependence-graph', 'figure'),
        [Input('dependence-color-col', 'value'),
         Input('dependence-highlight-index', 'value'),
         Input('label-store', 'data')],
        [State('dependence-col', 'value'),
         State('dependence-group-categoricals', 'checked')])
    def update_dependence_graph(color_col, idx, pos_label, col, cats):
        explainer.pos_label = pos_label #needed in case of multiple workers
        if color_col is not None:
            return explainer.plot_shap_dependence(
                        col, color_col, highlight_idx=idx, cats=cats)
        raise PreventUpdate


from abc import ABC, abstractmethod
import warnings
import base64

import numpy as np
import pandas as pd

from pdpbox import pdp
import shap
from dtreeviz.trees import *

from sklearn.metrics import roc_auc_score, accuracy_score, f1_score, precision_score, recall_score, log_loss
from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score

from .explainer_methods import *
from .explainer_plots import *


class BaseExplainerBunch(ABC):
    """Abstract Base Class. Defines the basic functionality of an ExplainerBunch
    But does not yet have a defined shap_explainer.
    """
    def __init__(self, model, X, y=None, metric=r2_score, shap="tree", 
                    cats=None, idxs=None, descriptions=None, permutation_cv=None, na_fill=-999):
        """init

        :param model: a model with a scikit-learn compatible .fit and .predict method
        :type model: [type]
        :param X: a pd.DataFrame with your model features
        :type X: pd.DataFrame
        :param y: Dependent variable of your model, defaults to None
        :type y: pd.Series, optional
        :param metric: is a scikit-learn compatible metric function (or string), 
            defaults to r2_score
        :type metric: metric function, optional
        :param shap: type of shap_explainer to fit, defaults to 'tree'
        :type shap: str
        :param cats: list of variables that have been onehotencoded. Allows to 
            group categorical variables together in plots, defaults to None
        :type cats: list, optional
        :param idxs: list of row identifiers. Can be names, id's, etc. Get 
            converted to str, defaults to None
        :type idxs: list, optional
        :param permutation_cv: If given permutation importances get calculated 
            using cross validation, defaults to None
        :type permutation_cv: int, optional
        :param na_fill: The filler used for missing values, defaults to -999
        :type na_fill: int, optional
        """
        self.model  = model
        self.X = X.reset_index(drop=True)
        if y is not None:
            self.y = pd.Series(y).reset_index(drop=True)
        else:
            self.y = pd.Series(np.full(len(X), np.nan))
        
        self.metric = metric
        self.shap = shap

        self.cats = cats
        if idxs is not None:
            if isinstance(idxs, list):
                self.idxs = [str(i) for i in idxs]
            else:
                self.idxs = list(idxs.astype(str))
        elif idxs=="use_index":
            self.idxs = list(X.index.astype(str))
        else:
            self.idxs = [str(i) for i in range(len(X))]
        self.descriptions = {} if descriptions is None else descriptions
        self.permutation_cv = permutation_cv
        self.na_fill = na_fill
        self.columns = self.X.columns.tolist()
        self.is_classifier = False
        self.is_regression = False

    @classmethod
    def from_ModelBunch(cls, model_bunch, raw_data, metric, shap='tree',
                        index_column=None, permutation_cv=None, na_fill=-999,
                        **kwargs):
        """create an ExplainerBunch from a ModelBunch. A ModelBunch is a class
        containing a model, data transformer, target name and used columns, with
        both a .transform() and a .predict() method.

        Assumes model is tree based.

        :param model_bunch: ModelBunch
        :type model_bunch: [type]
        :param raw_data: raw input data that will be transformed
        :type raw_data: pd.DataFrame
        :param metric: metric with which to evaluate permutation importances
        :type metric: metric function or str
        :param index_column: column that will be used for idxs, defaults to None
        :type index_column: str, optional
        :param permutation_cv: [description], defaults to None
        :type permutation_cv: [type], optional
        :param na_fill: [description], defaults to -999
        :type na_fill: int, optional
        :return: ExplainerBunch instance
        """
        assert not index_column or index_column in raw_data.columns, \
            f"{index_column} not in raw_data!"

        X, y = model_bunch.transform(raw_data)
        cats = raw_data.select_dtypes(include='object').columns.tolist()

        if index_column is not None:
            idxs = raw_data[index_column].astype(str).values.tolist()
        else:
            idxs = None
        return cls(model_bunch.model, X, y,  metric=metric, shap=shap,
                    cats=cats, idxs=idxs, permutation_cv=permutation_cv, na_fill=na_fill,
                    **kwargs)

    def __len__(self):
        return len(self.X)

    def __contains__(self, index):
        if self.get_int_idx(index) is not None:
            return True
        return False

    @property
    def shap_explainer(self):
        if not hasattr(self, '_shap_explainer'):
            print("Generating shap TreeExplainer...")
            if self.shap == 'tree':
                if str(type(self.model))[-15:-2]=='XGBClassifier':
                    warnings.warn("Warning: shap values for XGBoost models get calculated"
                        "against the background data X, not against the data the"
                        "model was trained on!")
                    self._shap_explainer = shap.TreeExplainer(
                                                self.model, self.X,
                                                model_output="probability",
                                                feature_dependence= "independent")
                else:
                    self._shap_explainer = shap.TreeExplainer(self.model)
            elif self.shap=='linear':
                self._shap_explainer = shap.LinearExplainer(self.model)
            elif self.shap=='deep':
                self._shap_explainer = shap.DeepExplainer(self.model)
            elif self.shap=='kernel': 
                self._shap_explainer = shap.KernelExplainer(self.model)
        return self._shap_explainer

    def get_int_idx(self, index):
        """
        Always returns an int index.
        If index is already int, simply return directly
        if index is str, lookup corresponding int index and return
        if index not found, return None
        """
        if isinstance(index, int):
            if index >= 0 and index < len(self):
                return index
        elif isinstance(index, str):
            if self.idxs is not None and index in self.idxs:
                return self.idxs.index(index)
        return None

    def random_index(self, y_min=None, y_max=None, 
                            pred_min=None, pred_max=None, return_str=False):
        """
        Return a random index from dataset.
        if y_values is given select an index for which y in y_values
        if return_str return str index from self.idxs
        """
        if y_min is None:
            y_min = self.y.min()
        if y_max is None:
            y_max = self.y.max()
        if pred_min is None:
            pred_min = self.preds.min()
        if pred_max is None:
            pred_max = self.preds.max()

        potential_idxs = self.y[(self.y>=y_min) & (self.y <= y_max) & 
                                (self.preds>=pred_min) & (self.preds <= pred_max)].index

        if len(potential_idxs) > 0:
            idx = np.random.choice(potential_idxs)
        else:
            return None
        if return_str:
            assert self.idxs is not None, \
                "no self.idxs property found..."
            return self.idxs[idx]
        return idx

    @property
    def preds(self):
        """returns model model predictions"""
        if not hasattr(self, '_preds'):
            print("Calculating predictions...")
            self._preds = self.model.predict(self.X)
            
        return self._preds

    @property
    def pred_percentiles(self):
        """returns percentile rank of model predictions""" 
        if not hasattr(self, '_pred_percentiles'):
            print("Calculating prediction percentiles...")
            self._pred_percentiles = (pd.Series(self.preds)
                                .rank(method='min')
                                .divide(len(self.preds))
                                .values)
        return self._pred_percentiles

    def columns_ranked_by_shap(self, cats=False):
        """returns the columns of X, ranked by mean abs shap value,
        - cats: if True group categorical features together""" 
        if cats:
            return self.mean_abs_shap_cats.Feature.tolist()
        else:
            return self.mean_abs_shap.Feature.tolist()

    def equivalent_col(self, col):
        """if col in self.columns, return equivalent col in self.columns_cats,
                e.g. equivalent_col('Gender_Male') -> 'Gender'
           if col in self.columns_cats, return first one hot encoded col, 
                e.g. equivalent_col('Gender') -> 'Gender_Male'

            (useful for switching between cats=True and cats=False, while
            maintaining column selection)
        """
        if col in self.columns_cats:
            # first onehot-encoded columns
            return get_feature_dict(self.columns, self.cats)[col][0]
        elif col in self.columns:
            # the cat that the col belongs to
            return [k for k, v in get_feature_dict(self.columns, self.cats).items() if col in v][0]
        return None

    def description(self, col):
        """returns the written out description of what feature col means."""
        if col in self.descriptions.keys():
            return self.descriptions[col]
        elif self.equivalent_col(col) in  self.descriptions.keys():
            return self.descriptions[self.equivalent_col(col)]
        return ""

    def description_list(self, cols):
        return [self.description(col) for col in cols]

    def get_col(self, col):
        """returns either the column values if col in self.columns,
        or induces the categorical values from the onehotencoding if col in self.cats
        """
        assert col in self.columns or col in self.cats, \
            f"{col} not in columns!"

        if col in self.X.columns:
            return self.X[col]
        elif col in self.cats:
            return retrieve_onehot_value(self.X, col)
        
    def get_col_value_plus_prediction(self, index, col):
        """return value of col and prediction for index
        helper function for get_pdp()"""
        assert index in self, f"index {index} not found"
        assert (col in self.X.columns) or (col in self.cats),\
            f"{col} not in columns of dataset"

        idx = self.get_int_idx(index)

        if col in self.X.columns:
            col_value = self.X[col].iloc[idx]
        elif col in self.cats:
            col_value = retrieve_onehot_value(self.X, col).iloc[idx]

        try:
            prediction = self.pred_probas[idx]
        except:
            prediction = self.preds[idx]

        return col_value, prediction

    @property
    def permutation_importances(self):
        """return the permatuation importances of the model features"""
        if not hasattr(self, '_perm_imps'):
            print("Calculating importances...")
            self._perm_imps = cv_permutation_importances(
                            self.model, self.X, self.y, self.metric,
                            cv=self.permutation_cv,
                            needs_proba=self.is_classifier)
        return self._perm_imps

    @property
    def permutation_importances_cats(self):
        """permutation importances with categoricals grouped"""
        if not hasattr(self, '_perm_imps_cats'):
            self._perm_imps_cats = cv_permutation_importances(
                            self.model, self.X, self.y, self.metric, self.cats,
                            cv=self.permutation_cv,
                            needs_proba=self.is_classifier)
        return self._perm_imps_cats

    @property
    def X_cats(self):
        """return model features DataFrame with onehot encoded categorical features
        reverse encoded to original state"""
        if not hasattr(self, '_X_cats'):
            self._X_cats = merge_categorical_columns(self.X, self.cats)
        return self._X_cats

    @property
    def columns_cats(self):
        """columns features with onehot encoded categorical features
        reverse encoded to original state"""
        if not hasattr(self, '_columns_cats'):
            self._columns_cats = self.X_cats.columns.tolist()
        return self._columns_cats

    @property
    def shap_base_value(self):
        """the intercept for the shap values. (i.e. 'what would the prediction be
        if we knew none of the features?')"""
        if not hasattr(self, '_shap_base_value'):
            self._shap_base_value = self.shap_explainer.expected_value
            if isinstance(self._shap_base_value, np.ndarray):
                # newer version of shap library returns an array instead of float
                self._shap_base_value = self._shap_base_value.item()
        return self._shap_base_value

    @property
    def shap_values(self):
        """SHAP values calculated using the shap library"""
        if not hasattr(self, '_shap_values'):
            print("Calculating shap values...")
            self._shap_values = self.shap_explainer.shap_values(self.X)
        return self._shap_values

    @property
    def shap_values_cats(self):
        """SHAP values when categorical features have been grouped"""
        if not hasattr(self, '_shap_values_cats'):
            print("Calculating shap values...")
            self._shap_values_cats = merge_categorical_shap_values(
                    self.X, self.shap_values, self.cats)
        return self._shap_values_cats

    @property
    def shap_interaction_values(self):
        """SHAP interaction values calculated using shap library"""
        if not hasattr(self, '_shap_interaction_values'):
            print("Calculating shap interaction values...")
            self._shap_interaction_values = \
                self.shap_explainer.shap_interaction_values(self.X)
        return self._shap_interaction_values

    @property
    def shap_interaction_values_cats(self):
        """SHAP interaction values with categorical features grouped"""
        if not hasattr(self, '_shap_interaction_values_cats'):
            print("Calculating shap interaction values...")
            self._shap_interaction_values_cats = \
                merge_categorical_shap_interaction_values(
                    self.X, self.X_cats, self.shap_interaction_values)
        return self._shap_interaction_values_cats

    @property
    def mean_abs_shap(self):
        """Mean absolute SHAP values per feature. Gives indication of overall
        importance of feature for predictions of model."""
        if not hasattr(self, '_mean_abs_shap'):
            self._mean_abs_shap = mean_absolute_shap_values(
                                self.columns, self.shap_values)
        return self._mean_abs_shap

    @property
    def mean_abs_shap_cats(self):
        """Mean absolute SHAP values per feature with categorical features grouped.
        Gives indication of overall importance of feature for predictions of model."""
        if not hasattr(self, '_mean_abs_shap_cats'):
            self._mean_abs_shap_cats = mean_absolute_shap_values(
                                self.columns, self.shap_values, self.cats)
        return self._mean_abs_shap_cats

    def calculate_properties(self, include_interactions=True):
        """Explicitely calculates all lazily calculated properties. Can be useful
        to call before saving ExplainerBunch to disk so that no need
        to calculate these properties when for example starting a Dashboard.

        :param include_interactions: shap interaction values can take a long time
        to compute for larger datasets with more features. Therefore you can choose
        not to calculate these, defaults to True
        :type include_interactions: bool, optional
        """
        _ = (self.preds, self.permutation_importances,
                self.shap_base_value, self.shap_values,
                self.shap_interaction_values, self.mean_abs_shap)
        if self.cats is not None:
            _ = (self.mean_abs_shap_cats, self.X_cats,
                    self.shap_values_cats)
        if include_interactions:
            _ = self.shap_interaction_values
            if self.cats is not None:
                _ = self.shap_interaction_values_cats

    @abstractmethod
    def metrics(self, **kwargs):
        """returns a dict of metrics. Implemented by either ClassifierBunch
        or RegressionBunch"""
        raise NotImplementedError()

    def metrics_markdown(self, round=2, **kwargs):
        """markdown makeup of self.metrics() dict"""
        metrics_dict = self.metrics(**kwargs)
        
        metrics_markdown = "# Model Summary: \n\n"
        for k, v in metrics_dict.items():
            metrics_markdown += f"### {k}: {np.round(v, round)}\n"
        return metrics_markdown
    
    def mean_abs_shap_df(self, topx=None, cutoff=None, cats=False):
        """returns a pd.DataFrame with the mean absolute shap values per features,
        sorted rom highest to lowest.

        :param topx: Only return topx most importance features, defaults to None
        :type topx: int, optional
        :param cutoff: Only return features with mean abs shap of at least cutoff, defaults to None
        :type cutoff: float, optional
        :param cats: group categorical variables, defaults to False
        :type cats: bool, optional
        :return:shap_df
        :rtype pd.DataFrame

        """
        shap_df = self.mean_abs_shap_cats if cats else self.mean_abs_shap

        if topx is None: topx = len(shap_df)
        if cutoff is None: cutoff = shap_df['MEAN_ABS_SHAP'].min()
        return (shap_df[shap_df['MEAN_ABS_SHAP'] >= cutoff]
                    .sort_values('MEAN_ABS_SHAP', ascending=False).head(topx))

    def shap_top_interactions(self, col, topx=None, cats=False):
        """returns the features that interact with feature col in descending order.

        if shap interaction values have already been calculated, use those. 
        Otherwise use shap approximate_interactions or simply shap importance.

        :param col: feature for which you want to get the interactions
        :type col: str
        :param topx: Only return topx features, defaults to None
        :type topx: int, optional
        :param cats: Group categorical features, defaults to False
        :type cats: bool, optional
        :return: top_interactions
        :rtype: list

        """
        if cats:
            if hasattr(self, '_shap_interaction_values'):
                col_idx = self.X_cats.columns.get_loc(col)
                top_interactions = self.X_cats.columns[np.argsort(-np.abs(
                        self.shap_interaction_values_cats[:, col_idx, :]).mean(0))].tolist()
            else:
                top_interactions = self.mean_abs_shap_cats.Feature.values.tolist()
                top_interactions.insert(0, top_interactions.pop(
                    top_interactions.index(col))) #put col first

            if topx is None: topx = len(top_interactions)
            return top_interactions[:topx]
        else:
            if hasattr(self, '_shap_interaction_values'):
                col_idx = self.X.columns.get_loc(col)
                top_interactions = self.X.columns[np.argsort(-np.abs(
                            self.shap_interaction_values[:, col_idx, :]).mean(0))].tolist()
            else:
                interaction_idxs = shap.common.approximate_interactions(
                    col, self.shap_values, self.X)
                top_interactions = self.X.columns[interaction_idxs].tolist()
                top_interactions.insert(0, top_interactions.pop(-1)) #put col first

            if topx is None: topx = len(top_interactions)
            return top_interactions[:topx]

    def shap_interaction_values_by_col(self, col, cats=False):
        """
        returns the shap interaction values[np.array(N,N)] for feature col

        :param col: features for which you'd like to get the interaction value
        :type col: str
        :param cats: group categorical, defaults to False
        :type cats: bool, optional
        :return: shap_interaction_values
        :rtype: np.array(N,N)
        """
        if cats:
            return self.shap_interaction_values_cats[:,
                        self.X_cats.columns.get_loc(col), :]
        else:
            return self.shap_interaction_values[:,
                        self.X.columns.get_loc(col), :]

    def permutation_importances_df(self, topx=None, cutoff=None, cats=False):
        """Returns pd.DataFrame with features ordered by permutation importance.
        For more about permutation importances see https://explained.ai/rf-importance/index.html

        :param topx: only return topx most important features, defaults to None
        :type topx: int, optional
        :param cutoff: only return features with importance of at least cutoff, defaults to None
        :type cutoff: float, optional
        :param cats: Group categoricals, defaults to False
        :type cats: bool, optional
        :return: importance_df
        :rtype: pd.DataFrame
        """
        importance_df = self.permutation_importances_cats.reset_index() if cats \
                                else self.permutation_importances.reset_index()

        if topx is None: topx = len(importance_df)
        if cutoff is None: cutoff = importance_df.Importance.min()
        return importance_df[importance_df.Importance > cutoff].head(topx)

    def importances_df(self, kind="shap", topx=None, cutoff=None, cats=False):
        """wrapper function for mean_abs_shap_df() and permutation_importance_df()"""
        assert kind=='shap' or kind=='permutation', "kind should either be 'shap' or 'permutation'!"
        if kind=='permutation':
            return self.permutation_importances_df(topx, cutoff, cats)
        elif kind=='shap':
            return self.mean_abs_shap_df(topx, cutoff, cats)

    def contrib_df(self, index, cats=True, topx=None, cutoff=None):
        """returns a contrib_df pd.DataFrame with the shap value contributions
        to the prediction for index. Used as input for the plot_contributions()
        method.

        :param index: index for which to calculate contributions
        :type index: int or str
        :param cats: Group categoricals, defaults to True
        :type cats: bool, optional
        :param topx: Only return topx features, remainder called REST, defaults to None
        :type topx: int, optional
        :param cutoff: only return features with at least cutoff contributions, defaults to None
        :type cutoff: float, optional
        :return: contrib_df
        :rtype: pd.DataFrame
        """
        idx = self.get_int_idx(index)
        if cats:
            return get_contrib_df(self.shap_base_value, self.shap_values_cats[idx],
                                    self.X_cats.iloc[[idx]], topx, cutoff)
        else:
            return get_contrib_df(self.shap_base_value, self.shap_values[idx],
                                    self.X.iloc[[idx]], topx, cutoff)

    def contrib_summary_df(self, index, cats=True,
                            topx=None, cutoff=None, round=2):
        """Takes a contrib_df, and formats it to a more human readable format"""
        idx = self.get_int_idx(index) # if passed str convert to int index
        return get_contrib_summary_df(self.contrib_df(idx, cats, topx, cutoff),
                                        classification=self.is_classifier,
                                        round=round)

    def interactions_df(self, col, cats=False, topx=None, cutoff=None):
        importance_df = mean_absolute_shap_values(
                            self.columns_cats if cats else self.columns, 
                            self.shap_interaction_values_by_col(col, cats))

        if topx is None: topx = len(importance_df)
        if cutoff is None: cutoff = importance_df.MEAN_ABS_SHAP.min()
        return importance_df[importance_df.MEAN_ABS_SHAP > cutoff].head(topx)
    
    def formatted_contrib_df(self, index, round=None, lang='en'):
        """Our PowerBI specialist wanted this the contrib_df in a certain format in order
        to conventiently build powerbi dashboards from the output of get_dfs.

        Additional language option for output in Dutch (lang='nl')

        :param index: index to return contrib_df for
        :type index: str or int
        :param round: rounding of continuous features, defaults to 2
        :type round: int, optional
        :param lang: language to name the columns, defaults to 'en'
        :type lang: str, optional
        :return: formatted_contrib_df
        :rtype: pd.DataFrame
        """
        cdf = self.contrib_df(index, cats=True).copy()
        cdf.reset_index(inplace=True)
        cdf.loc[cdf.col=='base_value', 'value'] = np.nan
        cdf['row_id'] = self.get_int_idx(index)
        cdf['name_id'] = self.idxs[self.get_int_idx(index)]
        cdf['cat_value'] = np.where(cdf.col.isin(self.cats), cdf.value, np.nan)
        cdf['cont_value'] = np.where(cdf.col.isin(self.cats), np.nan, cdf.value)
        if round is not None:
            rounded_cont = np.round(cdf['cont_value'].values.astype(float), round)
            cdf['value'] = np.where(cdf.col.isin(self.cats), cdf.cat_value, rounded_cont)
        cdf['type'] = np.where(cdf.col.isin(self.cats), 'cat', 'cont')
        cdf['abs_contribution'] = np.abs(cdf.contribution)
        cdf = cdf[['row_id', 'name_id', 'contribution', 'abs_contribution',
                    'col', 'value', 'cat_value', 'cont_value', 'type', 'index']]
        if lang == 'nl':
            cdf.columns = ['row_id', 'name_id', 'SHAP', 'ABS_SHAP', 'Variabele', 'Waarde',
                            'Cat_Waarde', 'Cont_Waarde', 'Waarde_Type', 'Variabele_Volgorde']
            return cdf

        cdf.columns = ['row_id', 'name_id', 'SHAP', 'ABS_SHAP', 'Feature', 'Value',
                        'Cat_Value', 'Cont_Value', 'Value_Type', 'Feature_Order']
        return cdf

    def get_pdp_result(self, col, index=None, drop_na=True,
                        sample=500, num_grid_points=20):
        """Uses the PDPBox to calculate partial dependences for feature col.

        :param col: Feature to calculate partial dependences for
        :type col: str
        :param index: Index of row to put at iloc[0], defaults to None
        :type index: int or str, optional
        :param drop_na: drop rows where col equals na_fill, defaults to True
        :type drop_na: bool, optional
        :param sample: Number of rows to sample for plot, defaults to 500
        :type sample: int, optional
        :param num_grid_points: Number of grid points to calculate, defaults to 20
        :type num_grid_points: int, optional
        :return: pdp_result
        :rtype: PDPBox.pdp_result
        """
        assert col in self.X.columns or col in self.cats, \
            f"{col} not in columns of dataset"

        features = get_feature_dict(self.X.columns, self.cats)[col]

        if index is not None:
            idx = self.get_int_idx(index)
            if len(features)==1 and drop_na: # regular col, not onehotencoded
                sample_size=min(sample, len(self.X[(self.X[features[0]] != self.na_fill)])-1)
                sampleX = pd.concat([
                    self.X[self.X.index==idx],
                    self.X[(self.X.index!=idx) & (self.X[features[0]] != self.na_fill)]\
                            .sample(sample_size)],
                    ignore_index=True, axis=0)
            else:
                sample_size = min(sample, len(self.X)-1)
                sampleX = pd.concat([
                    self.X[self.X.index==idx],
                    self.X[(self.X.index!=idx)].sample(sample_size)],
                    ignore_index=True, axis=0)
        else:
            if len(features)==1 and drop_na: # regular col, not onehotencoded
                sample_size=min(sample, len(self.X[(self.X[features[0]] != self.na_fill)])-1)
                sampleX = self.X[(self.X[features[0]] != self.na_fill)]\
                                .sample(sample_size)
            else:
                sampleX = self.X.sample(min(sample, len(self.X)))

        # if only a single value (i.e. not onehot encoded, take that value
        # instead of list):
        if len(features)==1: features=features[0]
        pdp_result = pdp.pdp_isolate(
                model=self.model, dataset=sampleX,
                model_features=self.X.columns,
                num_grid_points=num_grid_points, feature=features)
        if isinstance(features, list):
            # strip 'col_' from the grid points
            pdp_result.feature_grids = \
                pd.Series(pdp_result.feature_grids).str.split(col+'_').str[1].values
        return pdp_result

    def get_dfs(self, cats=True, round=None, lang='en'):
        """returns three pd.DataFrames. The first with id, prediction, actual and
        feature values, the second with only id and shap values. The third
        is similar to contrib_df for every id. 
        These can then be used to build your own custom dashboard on these data,
        for example using PowerBI.

        :param cats: group categorical variables, defaults to True
        :type cats: bool, optional
        :param round: how to round shap values
        :type round: int, optional
        :param lang: language to format dfs in. Defaults to 'en', 'nl' also available
        :type lang: str, optional
        :return: cols_df, shap_df, contribs_df
        :rtype: pd.DataFrame, pd.DataFrame, pd.DataFrame
        """
        if cats:
            cols_df = self.X_cats.copy()
            shap_df = pd.DataFrame(self.shap_values_cats, columns = self.X_cats.columns)
        else:
            cols_df = self.X.copy()
            shap_df = pd.DataFrame(self.shap_values, columns = self.X.columns)

        actual_str = 'Uitkomst' if lang == 'nl' else 'Actual'
        prediction_str = 'Voorspelling' if lang == 'nl' else 'Prediction'
        
        cols_df.insert(0, actual_str, self.y )
        if self.is_classifier:
            cols_df.insert(0, prediction_str, self.pred_probas)
        else:
            cols_df.insert(0, prediction_str, self.preds)
        cols_df.insert(0, 'name_id', self.idxs)
        cols_df.insert(0, 'row_id', range(len(self)))
 
        shap_df.insert(0, 'SHAP_base', np.repeat(self.shap_base_value, len(self)))
        shap_df.insert(0, 'name_id', self.idxs)
        shap_df.insert(0, 'row_id', range(len(self)))


        contribs_df = None
        for idx in range(len(self)):
            fcdf = self.formatted_contrib_df(idx, round=round, lang=lang)
            if contribs_df is None: contribs_df = fcdf
            else: contribs_df = pd.concat([contribs_df, fcdf])

        return cols_df, shap_df, contribs_df

    def to_sql(self, conn, schema, name, if_exists='replace',
                cats=True, round=None, lang='en'):
        """Writes three dataframes generated by .get_dfs() to a sql server.
        Tables will be called name_COLS and name_SHAP and name_CONTRBIB

        :param conn: database connecter acceptable for pd.to_sql
        :type conn: sqlalchemy.engine.Engine or sqlite3.Connection
        :param schema: schema to write to
        :type schema: str
        :param name: name prefix of tables
        :type name: str
        :param cats: group categorical variables, defaults to True
        :type cats: bool, optional
        :param if_exists: How to behave if the table already exists.
        :type if_exists: {‘fail’, ‘replace’, ‘append’}, default ‘fail’
        :param round: how to round shap values
        :type round: int, optional
        :param lang: language to format dfs in. Defaults to 'en', 'nl' also available
        :type lang: str, optional
        """
        cols_df, shap_df, contribs_df = self.get_dfs(cats, round, lang)
        cols_df.to_sql(con=conn, schema=schema, name=name+"_COLS",
                        if_exists=if_exists, index=False)
        shap_df.to_sql(con=conn, schema=schema, name=name+"_SHAP",
                        if_exists=if_exists, index=False)
        contribs_df.to_sql(con=conn, schema=schema, name=name+"_CONTRIB",
                        if_exists=if_exists, index=False)

    def plot_importances(self, kind='shap', topx=None, cats=False, round=3):
        """return Plotly fig with barchart of importances in descending order.

        :param type: 'shap' for mean absolute shap values, 'permutation' for
        permutation importances, defaults to 'shap'
        :type type: str, optional
        :param topx: Only return topx features, defaults to None
        :type topx: int, optional
        :param cats: Group categoricals defaults to False
        :type cats: bool, optional
        :return: fig
        :rtype: plotly.fig
        """
        importances_df = self.importances_df(kind=kind, topx=topx, cats=cats)
        if self.descriptions:
            descriptions = self.description_list(importances_df.Feature)
            return plotly_importances_plot(importances_df, descriptions, round=round)
        else:
            return plotly_importances_plot(importances_df, round=round)


    def plot_interactions(self, col, cats=False, topx=None):
        """return Plotly fig with mean absolute shap interaction value for column
        col. 

        :param col: column for which to generate shap interaction value
        :type type: str, optional
        :param cats: Group categoricals defaults to False
        :type cats: bool, optional
        :param topx: Only return topx features, defaults to None
        :type topx: int, optional
        :return: fig
        :rtype: plotly.fig
        """
        interactions_df = self.interactions_df(col, cats=cats, topx=topx)
        return plotly_importances_plot(interactions_df)

    def plot_shap_contributions(self, index, cats=True,
                                    topx=None, cutoff=None, round=2):
        """reutn Plotly fig with waterfall plot of shap value contributions
        to the model prediction for index.

        :param index: index for which to display prediction
        :type index: int or str
        :param cats: Group categoricals, defaults to True
        :type cats: bool, optional
        :param topx: Only display topx features, defaults to None
        :type topx: int, optional
        :param cutoff: Only display features with at least cutoff contribution, defaults to None
        :type cutoff: float, optional
        :param round: round contributions to round precision, defaults to 2
        :type round: int, optional
        :return: fig
        :rtype: plotly.Fig
        """
        contrib_df = self.contrib_df(self.get_int_idx(index), cats, topx, cutoff)
        return plotly_contribution_plot(contrib_df,
                    classification=self.is_classifier, round=round)

    def plot_shap_summary(self, topx=None, cats=False):
        """Displays all individual shap value for each feature in a horizontal
        scatter chart in descending order by mean absolute shap value.

        :param topx: Only display topx most important features , defaults to 10
        :type topx: int, optional
        :param cats: Group categoricals , defaults to False
        :type cats: bool, optional
        :return: fig
        :rtype: plotly.Fig
        """
        if cats:
            return plotly_shap_scatter_plot(
                                self.shap_values_cats,
                                self.X_cats,
                                self.importances_df(kind='shap', topx=topx, cats=True)\
                                        ['Feature'].values.tolist())
        else:
            return plotly_shap_scatter_plot(
                                self.shap_values,
                                self.X,
                                self.importances_df(kind='shap', topx=topx)\
                                        ['Feature'].values.tolist())

    def plot_shap_interaction_summary(self, col, topx=None, cats=False):
        """Displays all individual shap interaction values for each feature in a
        horizontal scatter chart in descending order by mean absolute shap value.

        :param col: feature for which
        :type col: [type]
        :param topx: [description], defaults to 10
        :type topx: int, optional
        :param cats: [description], defaults to False
        :type cats: bool, optional
        :return: [description]
        :rtype: [type]
        """
        interact_cols = self.shap_top_interactions(col, cats=cats)
        if topx is None: topx = len(interact_cols)
        if cats:
            return plotly_shap_scatter_plot(
                self.shap_interaction_values_by_col(col, cats=cats),
                self.X_cats, interact_cols[:topx])
        else:
            return plotly_shap_scatter_plot(
                self.shap_interaction_values_by_col(col),
                self.X, interact_cols[:topx])

    def plot_shap_dependence(self, col, color_col=None, highlight_idx=None, cats=False):
        """
        Plots a shap dependence plot:
            - on the x axis the possible values of the feature `col`
            - on the y axis the associated individual shap values

        :param color_col: if color_col provided then shap values colored (blue-red) according to feature color_col
        :param highlight_idx: individual observation to be highlighed in the plot.
        :param cats: group categorical variables
        """
        if cats:
            if col in self.cats:
                return plotly_shap_violin_plot(self.X_cats, self.shap_values_cats, col, color_col)
            else:
                return plotly_dependence_plot(self.X_cats, self.shap_values_cats,
                                                col, color_col,
                                                highlight_idx=highlight_idx,
                                                na_fill=self.na_fill)
        else:
            return plotly_dependence_plot(self.X, self.shap_values,
                                            col, color_col,
                                            highlight_idx=highlight_idx,
                                            na_fill=self.na_fill)

    def plot_shap_interaction(self, col, interact_col,
                                            highlight_idx=None, cats=False):
        """plots a dependence plot for shap interaction effects

        :param col: feature for which to find interaction values
        :type col: str
        :param interact_col: feature for which interaction value are displayed
        :type interact_col: str
        :param highlight_idx: idx that will be highlighted, defaults to None
        :type highlight_idx: int, optional
        :param cats: group categorical features, defaults to False
        :type cats: bool, optional
        :return: Plotly Fig
        :rtype: plotly.Fig
        """
        if cats and interact_col in self.cats:
            return plotly_shap_violin_plot(
                self.X_cats, 
                self.shap_interaction_values_by_col(col, cats),
                interact_col, col, interaction=True)
        else:
            return plotly_dependence_plot(self.X_cats if cats else self.X,
                self.shap_interaction_values_by_col(col, cats),
                interact_col, col, highlight_idx=highlight_idx,
                interaction=True)

    def plot_pdp(self, col, index=None, drop_na=True, sample=100,
                    num_grid_lines=100, num_grid_points=10):
        """returns plotly fig for a partial dependence plot showing ice lines
        for num_grid_lines rows, average pdp based on sample of sample.
        If index is given, display pdp for this specific index.

        :param col: feature to display pdp graph for
        :type col: str
        :param index: index to highlight in pdp graph, defaults to None
        :type index: int or str, optional
        :param drop_na: if true drop samples with value equal to na_fill, defaults to True
        :type drop_na: bool, optional
        :param sample: sample size on which the average pdp will be calculated, defaults to 100
        :type sample: int, optional
        :param num_grid_lines: number of ice lines to display, defaults to 100
        :type num_grid_lines: int, optional
        :param num_grid_points: number of points on the x axis to calculate the pdp for, defaults to 10
        :type num_grid_points: int, optional
        :return: fig
        :rtype: plotly.Fig
        """
        pdp_result = self.get_pdp_result(col, index,
                            drop_na=drop_na, sample=sample,
                            num_grid_points=num_grid_points)

        if index is not None:
            try:
                col_value, pred = self.get_col_value_plus_prediction(index, col)
                return plotly_pdp(pdp_result,
                                display_index=0, # the idx to be displayed is always set to the first row by self.get_pdp_result()
                                index_feature_value=col_value, index_prediction=pred,
                                feature_name=col,
                                num_grid_lines=min(num_grid_lines, sample, len(self.X)))
            except:
                return plotly_pdp(pdp_result, feature_name=col,
                        num_grid_lines=min(num_grid_lines, sample, len(self.X)))
        else:
            return plotly_pdp(pdp_result, feature_name=col,
                        num_grid_lines=min(num_grid_lines, sample, len(self.X)))


class RandomForestExplainerBunch(BaseExplainerBunch):
    """
    RandomForestBunch allows for the analysis of individual DecisionTrees that
    make up the RandomForest.
    """
    
    @property
    def graphviz_available(self):
        if not hasattr(self, '_graphviz_available'):
            try:
                import graphviz.backend as be
                cmd = ["dot", "-V"]
                stdout, stderr = be.run(cmd, capture_output=True, check=True, quiet=True)
            except:
                print("""
                WARNING: you don't seem to have graphviz in your path (cannot run 'dot -V'), 
                so no dtreeviz visualisation of decision trees will be shown on the shadow trees tab.

                See https://github.com/parrt/dtreeviz for info on how to properly install graphviz 
                for dtreeviz. 
                """)
                self._graphviz_available = False
            else:
                self._graphviz_available = True
        return self._graphviz_available

    @property
    def decision_trees(self):
        if not hasattr(self, '_decision_trees'):
            print("Generating ShadowDecTree for each individual decision tree...")
            self._decision_trees = get_decision_trees(self.model, self.X, self.y)
        return self._decision_trees

    def decisiontree_df(self, tree_idx, index):
        """returns a pd.DataFrame with all decision nodes of a particular
                        tree (indexed by tree_idx) for a particular observation
                        (indexed by index)"""
        assert tree_idx >= 0 and tree_idx < len(self.decision_trees), \
            f"tree index {tree_idx} outside 0 and number of trees ({len(self.decision_trees)}) range"
        idx=self.get_int_idx(index)
        assert idx >= 0 and idx < len(self.X), \
            f"=index {idx} outside 0 and size of X ({len(self.X)}) range"
        if self.is_classifier:
            return get_decisiontree_df(self.decision_trees[tree_idx], self.X.iloc[idx],
                    pos_label=self.pos_label)
        else:
            return get_decisiontree_df(self.decision_trees[tree_idx], self.X.iloc[idx])

    def decisiontree_df_summary(self, tree_idx, index, round=2):
        """formats decisiontree_df in a slightly more human readable format."""
        idx=self.get_int_idx(index)
        return decisiontree_df_summary(self.decisiontree_df(tree_idx, idx),
                    classifier=self.is_classifier, round=round)

    def decision_path_file(self, tree_idx, index):
        """get a dtreeviz visualization of a particular tree in the random forest. 
        - tree_idxs the n'th tree in the random forest
        - index for row index
        returns the path where the .svg file is stored."""
        if not self.graphviz_available:
            print("No graphviz 'dot' executable available!") 
            return None

        idx = self.get_int_idx(index)
        print(idx)
        if self.is_regression:
            viz = dtreeviz(self.model.estimators_[tree_idx],
               self.X, self.y, 
               target_name='Target',
               #orientation ='LR',  # left-right orientation
               feature_names=self.columns,
               X=self.X.iloc[idx, :],)
        elif self.is_classifier:
            viz = dtreeviz(self.model.estimators_[tree_idx],
               self.X, self.y, 
               target_name='Target',
               #orientation ='LR',  # left-right orientation
               feature_names=self.columns,
               class_names=self.labels,
               X=self.X.iloc[idx, :]) 
        return viz.save_svg()

    def decision_path(self, tree_idx, index):
        """get a dtreeviz visualization of a particular tree in the random forest. 
        - tree_idxs the n'th tree in the random forest
        - index for row index
        returns the a IPython display SVG object for e.g. jupyter notebook."""
        if not self.graphviz_available:
            print("No graphviz 'dot' executable available!") 
            return None

        from IPython.display import SVG
        svg_file = self.decision_path_file(tree_idx, index)
        return SVG(open(svg_file,'rb').read())

    def decision_path_encoded(self, tree_idx, index):
        """get a dtreeviz visualization of a particular tree in the random forest. 
        - tree_idxs the n'th tree in the random forest
        - index for row index
        returns the a base64 encoded image, for inclusion in websites (e.g. dashboard)
        """
        if not self.graphviz_available:
            print("No graphviz 'dot' executable available!")
            return None
        
        svg_file = self.decision_path_file(tree_idx, index)
        encoded = base64.b64encode(open(svg_file,'rb').read()) 
        #encoded = base64.b64encode(viz.svg()) 
        svg_encoded = 'data:image/svg+xml;base64,{}'.format(encoded.decode()) 
        return svg_encoded


    def plot_trees(self, index, highlight_tree=None, round=2):
        """returns a plotly barchart with the values of the predictions
                of each individual tree for observation idx
        """
        #print('explainer call')
        idx=self.get_int_idx(index)
        assert idx is not None, 'invalid index'
        if self.is_classifier:
            return plotly_tree_predictions(self.model, self.X.iloc[[idx]],
                        highlight_tree=highlight_tree, round=round, pos_label=self.pos_label)
        else:
            return plotly_tree_predictions(self.model, self.X.iloc[[idx]], 
                        highlight_tree=highlight_tree, round=round)

    def calculate_properties(self, include_interactions=True):
        _ = self.decision_trees
        super().calculate_properties(include_interactions)


class ClassifierBunch(BaseExplainerBunch):
    """
    ExplainerBunch for classification models. Defines the shap values for
    each possible class in the classifications.

    You assign the positive label class afterwards with e.g. .pos_label=0

    In addition defines a number of plots specific to classification problems
    such as a precision plot, confusion matrix, roc auc curve and pr auc curve.
    """
    def __init__(self, model,  X, y=None,  metric=roc_auc_score, shap='tree',
                    cats=None, idxs=None, descriptions=None,
                    permutation_cv=None, na_fill=-999,
                    labels=None, pos_label=1):
        """Compared to BaseExplainerBunch defines two additional parameters:
        :param labels: list of str labels for the different classes, defaults to e.g. ['0', '1'] for a binary classification
        :type labels: list of str, optional
        :param pos_label: class that should be used as the positive class, defaults to 1
        :type pos_label: int or str (if str, needs to be in labels), optional
        """
        super().__init__(model, X, y, metric, shap, cats, idxs, descriptions, permutation_cv, na_fill)

        if labels is not None:
            self.labels = labels
        elif hasattr(self.model, 'classes_'):
                self.labels = self.model.classes_
        else:
            self.labels = [str(i) for i in range(self.y.nunique())]
        self.pos_label = pos_label
        self.is_classifier = True

    @property
    def pos_label(self):
        return self._pos_label

    @pos_label.setter
    def pos_label(self, label):
        if isinstance(label, int) and label >=0 and label <len(self.labels):
            self._pos_label = label
        elif isinstance(label, str) and label in self.labels:
            self._pos_label = self.labels.index(label)
        else:
            raise ValueError(f"'{label}' not in labels")

    @property
    def pos_label_str(self):
        return self.labels[self.pos_label]

    @property
    def y_binary(self):
        return np.where(self.y.values==self.pos_label, 1, 0)

    def get_prop_for_label(self, prop:str, label):
        tmp = self.pos_label
        self.pos_label = label
        ret = getattr(self, prop)
        self.pos_label = tmp
        return ret

    @property
    def pred_probas(self):
        """returns pred_proba for pos_label class"""
        return self.pred_probas_raw[:, self.pos_label]

    @property
    def pred_percentiles(self):
        """returns ranks for pos_label class"""
        return self.pred_percentiles_raw[:, self.pos_label]

    @property
    def pred_probas_raw(self):
        """returns pred_probas with probability for each class"""
        if not hasattr(self, '_pred_probas'):
            print("Calculating prediction probabilities...")
            assert hasattr(self.model, 'predict_proba'), \
                "model does not have a predict_proba method!"
            self._pred_probas =  self.model.predict_proba(self.X)
        return self._pred_probas

    @property
    def pred_percentiles_raw(self):
        if not hasattr(self, '_pred_percentiles_raw'):
            print("Calculating pred_percentiles...")
            self._pred_percentiles_raw = (pd.DataFrame(self.pred_probas_raw)
                                .rank(method='min')
                                .divide(len(self.pred_probas_raw))
                                .values)
        return self._pred_percentiles_raw

    @property
    def permutation_importances(self):
        """return the permatuation importances of the model features"""
        if not hasattr(self, '_perm_imps'):
            print("Calculating importances...")
            self._perm_imps = [cv_permutation_importances(
                            self.model, self.X, self.y, self.metric,
                            cv=self.permutation_cv,
                            needs_proba=self.is_classifier,
                            pos_label=label) for label in range(len(self.labels))]
        return self._perm_imps[self.pos_label]

    @property
    def permutation_importances_cats(self):
        """permutation importances with categoricals grouped"""
        if not hasattr(self, '_perm_imps_cats'):
            self._perm_imps_cats = [cv_permutation_importances(
                            self.model, self.X, self.y, self.metric, self.cats,
                            cv=self.permutation_cv,
                            needs_proba=self.is_classifier,
                            pos_label=label) for label in range(len(self.labels))]
        return self._perm_imps_cats[self.pos_label]

    @property
    def shap_base_value(self):
        if not hasattr(self, '_shap_base_value'):
            self._shap_base_value = self.shap_explainer.expected_value
            if isinstance(self._shap_base_value, np.ndarray):
                self._shap_base_value = list(self._shap_base_value)
            if len(self.labels)==2 and isinstance(self._shap_base_value, (np.floating, float)):
                self._shap_base_value = [1-self._shap_base_value, self._shap_base_value]
            assert len(self._shap_base_value)==len(self.labels),\
                f"len(shap_explainer.expected_value)={len(self._shap_base_value)}"\
                 + "and len(labels)={len(self.labels)} do not match!"
        return self._shap_base_value[self.pos_label]

    @property
    def shap_values(self):
        if not hasattr(self, '_shap_values'):
            print("Calculating shap values...")
            self._shap_values = self.shap_explainer.shap_values(self.X)
            if not isinstance(self._shap_values, list) and len(self.labels)==2:
                self._shap_values = [1-self._shap_values, self._shap_values]

            assert len(self._shap_values)==len(self.labels),\
                f"len(shap_values)={len(self._shap_values)}"\
                 + f"and len(labels)={len(self.labels)} do not match!"
        return self._shap_values[self.pos_label]

    @property
    def shap_values_cats(self):
        if not hasattr(self, '_shap_values_cats'):
            _ = self.shap_values
            self._shap_values_cats = [
                merge_categorical_shap_values(
                    self.X, sv, self.cats) for sv in self._shap_values]
        return self._shap_values_cats[self.pos_label]

    @property
    def shap_interaction_values(self):
        if not hasattr(self, '_shap_interaction_values'):
            print("Calculating shap interaction values...")
            _ = self.shap_values #make sure shap values have been calculated
            self._shap_interaction_values = self.shap_explainer.shap_interaction_values(self.X)
            if not isinstance(self._shap_interaction_values, list) and len(self.labels)==2:
                self._shap_interaction_values = [1-self._shap_interaction_values,
                                                    self._shap_interaction_values]
            self._shap_interaction_values = [
                normalize_shap_interaction_values(siv, self.shap_values)
                    for siv, sv in zip(self._shap_interaction_values, self._shap_values)]
        return self._shap_interaction_values[self.pos_label]

    @property
    def shap_interaction_values_cats(self):
        if not hasattr(self, '_shap_interaction_values_cats'):
            _ = self.shap_interaction_values
            self._shap_interaction_values_cats = [
                merge_categorical_shap_interaction_values(
                    self.X, self.X_cats, siv) for siv in self._shap_interaction_values]
        return self._shap_interaction_values_cats[self.pos_label]

    @property
    def mean_abs_shap(self):
        if not hasattr(self, '_mean_abs_shap'):
            _ = self.shap_values
            self._mean_abs_shap = [mean_absolute_shap_values(
                                self.columns, sv) for sv in self._shap_values]
        return self._mean_abs_shap[self.pos_label]

    @property
    def mean_abs_shap_cats(self):
        if not hasattr(self, '_mean_abs_shap_cats'):
            _ = self.shap_values
            self._mean_abs_shap_cats = [mean_absolute_shap_values(
                                self.columns, sv, self.cats) for sv in self._shap_values]
        return self._mean_abs_shap_cats[self.pos_label]

    def cutoff_from_percentile(self, percentile, pos_label=None):
        if pos_label is None:
            return pd.Series(self.pred_probas).nlargest(int((1-percentile)*len(self))).min()
        else:
            return pd.Series(self.pred_probas_raw[:, pos_label]).nlargest(int((1-percentile)*len(self))).min()

    def metrics(self, cutoff=0.5):
        """returns a number of useful metrics for your classifier:
            accuracy, precision, recall, f1, roc auc, pr auc, log loss
            """
        metrics_dict = {
            'accuracy' : accuracy_score(self.y_binary, np.where(self.pred_probas > cutoff, 1, 0)),
            'precision' : precision_score(self.y_binary, np.where(self.pred_probas > cutoff, 1, 0)),
            'recall' : recall_score(self.y_binary, np.where(self.pred_probas > cutoff, 1, 0)),
            'f1' : f1_score(self.y_binary, np.where(self.pred_probas > cutoff, 1, 0)),
            'roc_auc_score' : roc_auc_score(self.y_binary, self.pred_probas),
            'pr_auc_score' : average_precision_score(self.y_binary, self.pred_probas),
            'log_loss' : log_loss(self.y_binary, self.pred_probas)
        }
        return metrics_dict

    def get_pdp_result(self, col, index=None, drop_na=True,
                        sample=1000, num_grid_points=20):
        """gets a the result out of the PDPBox library, adjust for multiple labels
        and returns it."""
        pdp_result = super().get_pdp_result(
                                col, index, drop_na, sample, num_grid_points)
        if len(self.labels)==2:
            # for binary classifer PDPBox only gives pdp for the positive class.
            # instead of a list of pdps for every class
            # so we simply inverse when predicting the negative class
            if self.pos_label==0:
                pdp_result.pdp = 1 - pdp_result.pdp
                pdp_result.ice_lines = 1-pdp_result.ice_lines
            return pdp_result
        else:
             return pdp_result[self.pos_label]

    def random_index(self, y_values=None, return_str=False,
                    pred_proba_min=None, pred_proba_max=None,
                    pred_percentile_min=None, pred_percentile_max=None):
        """
        Return a random index from dataset.
        if y_values is given select an index for which y in y_values
        if return_str return str index from self.idxs

        if pred_proba_min(max) is given, return an index with at least a predicted
        probabiity of positive class of pred_proba_min(max)

        if pred_percentile_min(max) is given, return an index with at least a predicted
        percentile of probabiity of positive class of pred_percentile_min(max)
        """
        if (y_values is None 
            and pred_proba_min is None and pred_proba_max is None
            and pred_percentile_min is None and pred_percentile_max is None):
            potential_idxs = self.y.index
        else:
            if y_values is None: y_values = self.y.unique().tolist()
            if not isinstance(y_values, list): y_values = [y_values]
            if pred_proba_min is None: pred_proba_min = self.pred_probas.min()
            if pred_proba_max is None: pred_proba_max = self.pred_probas.max()
            if pred_percentile_min is None: pred_percentile_min = 0.0
            if pred_percentile_max is None: pred_percentile_max = 1.0
            
            potential_idxs = self.y[(self.y.isin(y_values)) &
                            (self.pred_probas >= pred_proba_min) &
                            (self.pred_probas <= pred_proba_max) &
                            (self.pred_percentiles > pred_percentile_min) &
                            (self.pred_percentiles <= pred_percentile_max)].index
        if not potential_idxs.empty:
            idx = np.random.choice(potential_idxs)
        else:
            return None
        if return_str:
            assert self.idxs is not None, \
                "no self.idxs property found..."
            return self.idxs[idx]
        return idx

    def precision_df(self, bin_size=None, quantiles=None, multiclass=False):
        """returns a pd.DataFrame with predicted probabilities and actually
        observed number of positive cases (i.e. precision)

        :param bin_size: group predictions in bins of size bin_size, defaults to 0.1
        :type bin_size: float, optional
        :param quantiles: group predictions in evenly sized quantiles of size quantiles, defaults to None
        :type quantiles: int, optional
        :param multiclass: whether to calculate precision for every class
        :type multiclass: bool, optional
        :return: precision_df
        :rtype: pd.DataFrame
        """
        assert self.pred_probas is not None
        if bin_size is None and quantiles is None:
            bin_size=0.1 # defaults to bin_size=0.1
        if multiclass:
            return get_precision_df(self.pred_probas_raw, self.y,
                                bin_size, quantiles, pos_label=self.pos_label)
        else:
            return get_precision_df(self.pred_probas, self.y_binary, bin_size, quantiles)

    def lift_curve_df(self):
        """returns a pd.DataFrame with data needed to build a lift curve"""
        return get_lift_curve_df(self.pred_probas, self.y, self.pos_label)

    def prediction_result_markdown(self, index, include_percentile=True, round=2, **kwargs):
        """returns a string with markdown syntax with the essential result of a prediction
        for a particular index.
        
        :param index: the index of the row for which to generate the prediction
        :type index: int or str
        :param include_percentile: include the rank percentile of the prediction, defaults to True
        :type include_percentile: bool, optional
        :param round: rounding to apply to results, defaults to 2
        :type round: int, optional
        :return: markdown string
        :rtype: str
        """
        int_idx = self.get_int_idx(index)
        
        def display_probas(pred_probas_raw, labels, round=2):
            assert (len(pred_probas_raw.shape)==1 
                    and len(pred_probas_raw) ==len(labels))
            for i in range(len(labels)):
                yield '##### ' + labels[i] + ': ' + str(np.round(100*pred_probas_raw[i], round)) + '%\n'

        model_prediction = f"# Prediction for {index}:\n" 
        for pred in display_probas(
                self.pred_probas_raw[int_idx], 
                self.labels, round):
            model_prediction += pred
        if (isinstance(self.y[0], int) or 
            isinstance(self.y[0], np.int64)):
            model_prediction += f"##### Actual Outcome: {self.labels[self.y[int_idx]]}\n\n"
        if include_percentile:
            model_prediction += f'##### In top {np.round(100*(1-self.pred_percentiles[int_idx]))}% percentile probability {self.pos_label_str}'
        return model_prediction

    def plot_precision(self, bin_size=None, quantiles=None, cutoff=0.5, multiclass=False):
        """plots predicted probability on the x-axis and observed precision (fraction of actual positive
        cases) on the y-axis.

        Should pass either bin_size fraction of number of quantiles, but not both. 


        :param bin_size: size of the bins on x-axis (e.g. 0.05 for 20 bins)
        :type bin_size: float
        :quantiles: number of equal sized quantiles to split the predictions by e.g. 20
        :type quantiles: int
        :param cutoff: cutoff of model to include in the plot
        :param multiclass: whether to display all classes or only positive class, defaults to False
        :return: Plotly fig
        """

        if bin_size is None and quantiles is None:
            bin_size=0.1 # defaults to bin_size=0.1
        precision_df = self.precision_df(
                bin_size=bin_size, quantiles=quantiles, multiclass=multiclass)
        return plotly_precision_plot(precision_df,
                    cutoff=cutoff, labels=self.labels, pos_label=self.pos_label)

    def plot_cumulative_precision(self):
        """returns a cumulative precision plot, which is a slightly different 
        representation of a lift curve.
        
        :return: plotly fig
        """
        return plotly_cumulative_precision_plot(
                    self.lift_curve_df(), self.labels, self.pos_label)

    def plot_confusion_matrix(self, cutoff=0.5, normalized=False, binary=False):
        """returns a plot of a confusion matrix.
        
        :param cutoff: cutoff of positive class to calculate confusion matrix for, defaults to 0.5
        :type cutoff: float, optional
        :param normalized: display percentages instead of counts , defaults to False
        :type normalized: bool, optional
        :param binary: if multiclass display one-vs-rest instead, defaults to False
        :type binary: bool, optional
        :return: [description]
        :rtype: [type]
        """
        
        if binary:
            if len(self.labels)==2:
                def order_binary_labels(labels, pos_label):
                    pos_index = labels.index(pos_label)
                    return [labels[1-pos_index], labels[pos_index]]
                labels = order_binary_labels(self.labels, self.pos_label_str)
            else:
                labels = ['Not ' + self.pos_label_str, self.pos_label_str]

            return plotly_confusion_matrix(
                    self.y_binary, np.where(self.pred_probas > cutoff, 1, 0),
                    normalized=normalized, labels=labels)
        else:
            return plotly_confusion_matrix(
                self.y, self.pred_probas_raw.argmax(axis=1),
                normalized=normalized, labels=self.labels)

    def plot_lift_curve(self, cutoff=None, percentage=False, round=2):
        return plotly_lift_curve(self.lift_curve_df(), cutoff, percentage, round)

    def plot_cumulative_precision(self):
        return plotly_cumulative_precision_plot(self.lift_curve_df(), 
                labels=self.labels, pos_label=self.pos_label)

    def plot_classification(self, cutoff=0.5, percentage=True):
        return plotly_classification_plot(self.pred_probas, self.y, self.labels, cutoff, percentage=percentage)

    def plot_roc_auc(self, cutoff=0.5):
        """plots ROC_AUC curve. The TPR and FPR of a particular
            cutoff is displayed in crosshairs."""
        return plotly_roc_auc_curve(self.y_binary, self.pred_probas, cutoff=cutoff)

    def plot_pr_auc(self, cutoff=0.5):
        """plots PR_AUC curve. the precision and recall of particular
            cutoff is displayed in crosshairs."""
        return plotly_pr_auc_curve(self.y_binary, self.pred_probas, cutoff=cutoff)

    def calculate_properties(self, include_interactions=True):
        _ = self.pred_probas
        super().calculate_properties(include_interactions)


class RegressionBunch(BaseExplainerBunch):
    """
     ExplainerBunch for regression models.

    In addition defines a number of plots specific to regression problems
    such as a predicted vs actual and residual plots.
    """
    def __init__(self, model,  X, y=None, metric=roc_auc_score, shap="tree",
                    cats=None, idxs=None, descriptions=None, 
                    permutation_cv=None, na_fill=-999,
                    units=""):
        """Combared to BaseExplainerBunch defines two additional parameters:
        :param units: units to display for regression quantity
        :type units: str, optional

        """
        super().__init__(model, X, y, metric, shap, cats, idxs, descriptions, permutation_cv, na_fill)
        self.units = units
        self.is_regression = True
    
    @property
    def residuals(self):
        if not hasattr(self, '_residuals'):
            print("Calculating residuals...")
            self._residuals =  self.preds-self.y
        return self._residuals

    def prediction_result_markdown(self, index, round=2, **kwargs):
        int_idx = self.get_int_idx(index)

        model_prediction = f"# Prediction for {index}:\n" \
                            + f"##### Prediction: {np.round(self.preds[int_idx], round)}\n"
        model_prediction += f"##### Actual Outcome: {np.round(self.y[int_idx], round)}"
        return model_prediction

    def metrics(self):
        """return dict of regression relevant metrics: rmse, mae and R2"""
        metrics_dict = {
            'rmse' : np.sqrt(mean_squared_error(self.y, self.preds)),
            'mae' : mean_absolute_error(self.y, self.preds),
            'R2' : r2_score(self.y, self.preds),
        }
        return metrics_dict

    def plot_predicted_vs_actual(self, round=2, logs=False):
        """returns a plot with predicted value on x-axis and actual value on y axis.
        
        :param round: rounding to apply to outcome, defaults to 2
        :type round: int, optional
        :param logs: whether to take logs of predicted and actual value, defaults to False
        :type logs: bool, optional
        :return: Plotly fig
        """
        return plotly_predicted_vs_actual(self.y, self.preds, units=self.units, round=round, logs=logs)
    
    def plot_residuals(self, vs_actual=False, round=2, ratio=False):
        """returns a plot of residuals. x-axis is the predicted outcome by default
        
        :param vs_actual: use actual value for x-axis, defaults to False
        :type vs_actual: bool, optional
        :param round: rounding to perform on values, defaults to 2
        :type round: int, optional
        :param ratio: whether to take the residual/prediction ratio instead, defaults to False
        :type ratio: bool, optional
        :return: Plotly fig
        """
        return plotly_plot_residuals(self.y, self.preds, 
                                     vs_actual=vs_actual, units=self.units, round=round, ratio=ratio)
    
    def plot_residuals_vs_feature(self, col, ratio=False, round=2, dropna=True):
        """Plot residuals vs individual features to see if certain values of
        features have higher residuals than others.
        
        :param col: Plot against feature col	
        :type col: str
        :param ratio: display residual ratio instead of raw value, defaults to False
        :type ratio: bool, optional
        :param round: rounding to perform on residuals, defaults to 2
        :type round: int, optional
        :param dropna: trop missing values from plot, defaults to True
        :type dropna: bool, optional
        :return: plotly fig
        """
        assert col in self.columns, \
            f'{col} not in columns!'
        na_mask = self.X[col] != self.na_fill if dropna else np.array([True]*len(self.X))
        return plotly_residuals_vs_col(self.y[na_mask], self.preds[na_mask], self.X[col][na_mask], 
                                       ratio=ratio, units=self.units, round=round)
    

class RandomForestClassifierBunch(RandomForestExplainerBunch, ClassifierBunch):
    """RandomForestClassifierBunch inherits from both RandomForestBunch and
    ClassifierBunch.
    """


class RandomForestRegressionBunch(RandomForestExplainerBunch, RegressionBunch):
    """RandomForestClassifierBunch inherits from both RandomForestBunch and
    RegressionBunch.
    """







from functools import partial

import numpy as np
import pandas as pd

from dtreeviz.trees import ShadowDecTree

from sklearn.metrics import make_scorer
from sklearn.base import clone
from sklearn.model_selection import StratifiedKFold


def get_feature_dict(all_cols, cats=None):
    """
    This helper function makes it easy to loop though columns in a dataframe
    and group onehot encoded columns together.

    :param all_cols: all columns of a dataframe
    :type all_cols: list
    :param cats: categorical columns that have been onehotencoded, defaults to None
    :type cats: list, optional
    :return: returns a dict with as key all original (not one hot encoded) columns
        and as items a list of columns associated with that column.

        e.g. {'Age': ['Age'],
              'Gender' : ['Gender_Male', 'Gender_Female']}
    :rtype: dict
    """
    feature_dict = {}

    if cats is None:
        return {col: [col] for col in all_cols}

    for col in cats:
        cat_cols = [c for c in all_cols if c.startswith(col)]
        if len(cat_cols) > 1:
            feature_dict[col] = cat_cols

    # add all the individual features
    other_cols = list(set(all_cols) - set([item for sublist in list(feature_dict.values())
                                                for item in sublist]))

    for col in other_cols:
        feature_dict[col] = [col]
    return feature_dict


def retrieve_onehot_value(X, encoded_col):
    """
    Returns a pd.Series with the original values that were onehot encoded.

    i.e. Finds the column name starting with encoded_col_ that has a value of 1.
        if no such column exists (they are all 0), then return 'NOT_ENCODED'
    """
    cat_cols = [c for c in X.columns if c.startswith(encoded_col + '_')]

    assert len(cat_cols) > 0, \
        f"No columns that start with {encoded_col} in DataFrame"

    feature_value = np.argmax(X[cat_cols].values, axis=1)

    # if not a single 1 then encoded feature must have been dropped
    feature_value[np.max(X[cat_cols].values, axis=1) == 0] = -1
    mapping = {-1: "NOT_ENCODED"}
    mapping.update({i: col[len(encoded_col) + 1:] for i, col in enumerate(cat_cols)})

    return pd.Series(feature_value).map(mapping)


def merge_categorical_columns(X, cats=None):
    """
    Returns a new feature Dataframe X_cats where the onehotencoded
    categorical features have been merged back with the old value retrieved
    from the encodings.
    """
    feature_dict = get_feature_dict(X.columns, cats)
    X_cats = X.copy()
    for col_name, col_list in feature_dict.items():
        if len(col_list) > 1:
            X_cats[col_name] = retrieve_onehot_value(X, col_name)
            X_cats.drop(col_list, axis=1, inplace=True)
    return X_cats

def merge_categorical_shap_values(X, shap_values, cats=None):
    """
    Returns a new feature Dataframe X_cats and new shap values np.array
    where the shap values of onehotencoded categorical features have been
    added up.
    """
    feature_dict = get_feature_dict(X.columns, cats)
    shap_df = pd.DataFrame(shap_values, columns=X.columns)
    for col_name, col_list in feature_dict.items():
        if len(col_list) > 1:
            shap_df[col_name] = shap_df[col_list].sum(axis=1)
            shap_df.drop(col_list, axis=1, inplace=True)
    return shap_df.values


def merge_categorical_shap_interaction_values(old_columns, new_columns,
                                                shap_interaction_values):
    """
    Returns a 3d numpy array shap_interaction_values where the categorical
    columns have been added up.

    Caution:
    Column names in new_columns that are not found in old_columns are
    assumed to be categorical feature names.
    """

    if isinstance(old_columns, pd.DataFrame):
        old_columns = old_columns.columns.tolist()
    if isinstance(new_columns, pd.DataFrame):
        new_columns = new_columns.columns.tolist()

    if not isinstance(old_columns, list):
        old_columns = old_columns.tolist()
    if not isinstance(new_columns, list):
        new_columns = new_columns.tolist()

    cats = [col for col in new_columns if col not in old_columns]
    feature_dict = get_feature_dict(old_columns, cats)

    siv = np.zeros((shap_interaction_values.shape[0],
                    len(new_columns),
                    len(new_columns)))

    for new_col1 in new_columns:
        for new_col2 in new_columns:
            newcol_idx1 = new_columns.index(new_col1)
            newcol_idx2 = new_columns.index(new_col2)
            oldcol_idxs1 = [old_columns.index(col)
                                for col in feature_dict[new_col1]]
            oldcol_idxs2 = [old_columns.index(col)
                                for col in feature_dict[new_col2]]
            siv[:, newcol_idx1, newcol_idx2] = \
                shap_interaction_values[:, oldcol_idxs1, :][:, :, oldcol_idxs2]\
                .sum(axis=(1, 2))

    return siv


def make_one_vs_all_scorer(metric, pos_label, greater_is_better=True):
    """
    Returns a binary one vs all scorer for a single class('pos_label') of a
    multiclass classifier
    """
    def one_vs_all_metric(metric, pos_label, y_true, y_pred):
        return metric((y_true == pos_label).astype(int), y_pred[:, pos_label])
    partial_metric = partial(one_vs_all_metric, metric, pos_label)
    sign = 1 if greater_is_better else -1

    def _scorer(clf, X, y):
        y_pred = clf.predict_proba(X)
        score = sign * partial_metric(y, y_pred)
        return score

    return _scorer


def permutation_importances(model, X, y, metric, cats=None,
                            greater_is_better=True, needs_proba=False,
                            pos_label=None,
                            sort=True, verbose=0):
    """
    adapted from rfpimp
    """
    X = X.copy()

    feature_dict = get_feature_dict(X.columns, cats)

    if isinstance(metric, str):
        scorer = make_scorer(metric, greater_is_better, needs_proba)
    elif not needs_proba or pos_label is None:
        scorer = make_scorer(metric, greater_is_better, needs_proba)
    else:
        scorer = make_one_vs_all_scorer(metric, pos_label, greater_is_better)

    baseline = scorer(model, X, y)
    imp = pd.DataFrame({'Importance': []})

    for col_name, col_list in feature_dict.items():
        old_cols = X[col_list].copy()
        X[col_list] = np.random.permutation(X[col_list])

        permutation_score = scorer(model, X, y)

        drop_in_metric = baseline - permutation_score
        imp = imp.append(pd.DataFrame({'Importance': [drop_in_metric]}, index=[col_name]))
        X[col_list] = old_cols

    imp.index.name = 'Feature'
    if sort:
        return imp.sort_values('Importance', ascending=False)
    else:
        return imp


def cv_permutation_importances(model, X, y, metric, cats=None, greater_is_better=True,
                                needs_proba=False, pos_label=None, cv=None, verbose=0):
    """
    Returns the permutation importances averages over `cv` cross-validated folds.
    """
    if cv is None:
        return permutation_importances(model, X, y, metric, cats,
                                        greater_is_better=greater_is_better,
                                        needs_proba=needs_proba,
                                        pos_label=pos_label,
                                        sort=False,
                                        verbose=verbose)

    skf = StratifiedKFold(n_splits=cv, random_state=None, shuffle=False)
    model = clone(model)
    for i, (train_index, test_index) in enumerate(skf.split(X, y)):
        X_train, X_test = X.iloc[train_index], X.iloc[test_index]
        y_train, y_test = y.iloc[train_index], y.iloc[test_index]

        model.fit(X_train, y_train)

        imp = permutation_importances(model, X_test, y_test, metric, cats,
                                        greater_is_better=greater_is_better,
                                        needs_proba=needs_proba,
                                        pos_label=pos_label,
                                        sort=False,
                                        verbose=verbose)
        if i == 0:
            imps = imp
        else:
            imps = imps.merge(imp, on='Feature', suffixes=("", "_" + str(i)))

    return pd.DataFrame(imps.mean(axis=1), columns=['Importance'])\
                        .sort_values('Importance', ascending=False)


def mean_absolute_shap_values(columns, shap_values, cats=None):
    """
    Returns a dataframe with the mean absolute shap values for each feature.
    """
    feature_dict = get_feature_dict(columns, cats)

    shap_abs_mean_dict = {}
    for col_name, col_list in feature_dict.items():
        shap_abs_mean_dict[col_name] = np.absolute(
            shap_values[:, [columns.index(col) for col in col_list]].sum(axis=1)
        ).mean()

    shap_df = pd.DataFrame(
        {
            'Feature': list(shap_abs_mean_dict.keys()),
            'MEAN_ABS_SHAP': list(shap_abs_mean_dict.values())
        }).sort_values('MEAN_ABS_SHAP', ascending=False).reset_index(drop=True)
    return shap_df


def get_precision_df(pred_probas, y_true, bin_size=None, quantiles=None, pos_label=1):
    """
    returns a pd.DataFrame with the predicted probabilities and
    the observed frequency per bin_size.

    If pred_probas has one dimension (i.e. only probabilities of positive class)
    only returns a single precision. If pred_probas containts probabilities for
    every class (typically a multiclass classifier), also returns precision
    for every class in every bin.
    """
    if bin_size is None and quantiles is None:
        bin_size = 0.1

    assert ((bin_size is not None and quantiles is None)
            or (bin_size is None and quantiles is not None)), \
        "either only pass bin_size or only pass quantiles!"

    
    if len(pred_probas.shape) == 2:
        # in case the full binary classifier pred_proba is passed,
        # we only select the probability of the positive class
        predictions_df = pd.DataFrame(
            {'pred_proba': pred_probas[:, pos_label], 'target': y_true})
        n_classes = pred_probas.shape[1]
    else:
        predictions_df = pd.DataFrame(
            {'pred_proba': pred_probas, 'target': y_true})
        n_classes = 1
        
    predictions_df = predictions_df.sort_values('pred_proba')

    # define a placeholder df:
    columns = ['p_min', 'p_max', 'p_avg', 'bin_width', 'precision', 'count']
    if n_classes > 1:
        for i in range(n_classes):
            columns.append('precision_' + str(i))

    precision_df = pd.DataFrame(columns=columns)

    if bin_size:
        thresholds = np.arange(0.0, 1.0, bin_size).tolist()
        # loop through prediction intervals, and compute
        for bin_min, bin_max in zip(thresholds, thresholds[1:] + [1.0]):
            if bin_min != bin_max:
                new_row_dict = {
                    'p_min': [bin_min],
                    'p_max': [bin_max],
                    'p_avg': [bin_min + (bin_max - bin_min) / 2.0],
                    'bin_width': [bin_max - bin_min]
                }

                if bin_min == 0.0:
                    new_row_dict['p_avg'] = predictions_df[
                            (predictions_df.pred_proba >= bin_min)
                            & (predictions_df.pred_proba <= bin_max)
                        ]['pred_proba'].mean()
                    new_row_dict['precision'] = (
                        predictions_df[
                            (predictions_df.pred_proba >= bin_min)
                            & (predictions_df.pred_proba <= bin_max)
                        ].target == pos_label
                    ).mean()
                    new_row_dict['count'] = predictions_df[
                            (predictions_df.pred_proba >= bin_min)
                            & (predictions_df.pred_proba <= bin_max)
                        ].target.count()
                    if n_classes > 1:
                        for i in range(n_classes):
                            new_row_dict['precision_' + str(i)] = (
                                predictions_df[
                                    (predictions_df.pred_proba >= bin_min)
                                    & (predictions_df.pred_proba <= bin_max)
                                ].target == i
                            ).mean()
                else:
                    new_row_dict['p_avg'] = predictions_df[
                            (predictions_df.pred_proba > bin_min)
                            & (predictions_df.pred_proba <= bin_max)
                        ]['pred_proba'].mean()
                    new_row_dict['precision'] = (
                        predictions_df[
                            (predictions_df.pred_proba > bin_min)
                            & (predictions_df.pred_proba <= bin_max)
                        ].target == pos_label
                    ).mean()
                    new_row_dict['count'] = (
                        predictions_df[
                            (predictions_df.pred_proba > bin_min)
                            & (predictions_df.pred_proba <= bin_max)
                        ].target == pos_label
                    ).count()
                    if n_classes > 1:
                        for i in range(n_classes):
                            new_row_dict['precision_' + str(i)] = (
                                predictions_df[
                                    (predictions_df.pred_proba > bin_min)
                                    & (predictions_df.pred_proba <= bin_max)
                                ].target == i
                            ).mean()
                new_row_df = pd.DataFrame(new_row_dict, columns=precision_df.columns)
                precision_df = pd.concat([precision_df, new_row_df])
        
    elif quantiles:
        preds_quantiles = np.array_split(predictions_df.pred_proba.values, quantiles)
        target_quantiles = np.array_split(predictions_df.target.values, quantiles)

        last_p_max = 0.0
        for preds, targets in zip(preds_quantiles, target_quantiles):
            new_row_dict = {
                    'p_min': [last_p_max],
                    'p_max': [preds.max()],
                    'p_avg': [preds.mean()],
                    'bin_width': [preds.max() - last_p_max],
                    'precision': [np.mean(targets==pos_label)],
                    'count' : [len(preds)],

                }
            if n_classes > 1:
                for i in range(n_classes):
                    new_row_dict['precision_' + str(i)] = np.mean(targets==i)

            new_row_df = pd.DataFrame(new_row_dict, columns=precision_df.columns)
            precision_df = pd.concat([precision_df, new_row_df])
            last_p_max = preds.max()
    
    return precision_df


def get_lift_curve_df(pred_probas, y, pos_label=1):
    """returns a pd.DataFrame that can be used to generate a lift curve plot.
    
    :param pred_probas: predicted probabilities of the positive class
    :type pred_probas: pd.Series, list of np.array
    :param y: the actual outcomes, encoded 0, 1 [, 2, 3, etc]
    :type y: pd.Series, list of np.array
    :param pos_label: the label of the positive class [defaults to 1]
    :type pos_label: int
    :return: lift_df
    :rtype: pd.DataFrame
    """
    lift_df = pd.DataFrame(
        {
            'pred_proba' : pred_probas, 
             'y' : y
        }).sort_values('pred_proba', ascending=False).reset_index(drop=True)
    lift_df['index'] = lift_df.index + 1
    lift_df['index_percentage'] = 100*lift_df['index'] / len(lift_df)
    lift_df['positives'] = (lift_df.y==pos_label).astype(int).cumsum()
    lift_df['precision'] = 100 * (lift_df['positives'] /  lift_df['index'])
    lift_df['cumulative_percentage_pos'] = 100 * (lift_df['positives'] / (lift_df.y==pos_label).astype(int).sum())
    lift_df['random_pos'] = (lift_df.y==pos_label).astype(int).mean() * lift_df['index']
    lift_df['random_precision'] = 100 * (lift_df['random_pos'] /  lift_df['index'])
    lift_df['random_cumulative_percentage_pos'] = 100 * (lift_df['random_pos'] / (lift_df.y==pos_label).astype(int).sum())
    for y_label in range(y.nunique()):
        lift_df['precision_' + str(y_label)] = 100*(lift_df.y==y_label).astype(int).cumsum() / lift_df['index']
    return lift_df
    

def get_contrib_df(shap_base_value, shap_values, X_row, topx=None, cutoff=None):
    """
    Return a contrib_df DataFrame that lists the contribution of each input
    variable.

    X_row should be a single row of features, generated using X.iloc[[index]]
    if topx is given, only returns the highest topx contributions
    if cutoff is given only returns contributions above cutoff.
    """
    assert isinstance(X_row, pd.DataFrame),\
        'X_row should be a pd.DataFrame! Use X.iloc[[index]]'
    assert len(X_row.iloc[[0]].values[0].shape) == 1,\
        """X is not the right shape: len(X.values[0]) should be 1. 
            Try passing X.iloc[[index]]""" 

    # start with the shap_base_value
    base_df = pd.DataFrame(
        {
            'col': ['base_value'],
            'contribution': [shap_base_value],
            'value': ['-']
        })

    contrib_df = pd.DataFrame(
                    {
                        'col': X_row.columns,
                        'contribution': shap_values,
                        'value': X_row.values[0]
                    })

    # sort the df by absolute value from highest to lowest:
    contrib_df = contrib_df.reset_index(drop=True)
    contrib_df = contrib_df.reindex(
                    contrib_df.contribution.abs().sort_values(ascending=False).index)

    contrib_df = pd.concat([base_df, contrib_df], ignore_index=True)

    # add cumulative contribution from top to bottom (for making graph):
    contrib_df['cumulative'] = contrib_df.contribution.cumsum()

    # if a cutoff is given for minimum contribution to be displayed, calculate what topx rows to return:
    if cutoff is not None:
        cutoff = contrib_df.contribution[np.abs(contrib_df.contribution) >= cutoff].index.max()+1
        if topx is not None and cutoff < topx:
            topx = cutoff

    # if only returning topx columns, sum the remainder contributions under 'REST'
    if topx is not None:
        if topx > len(contrib_df): topx = len(contrib_df)
        old_cum = contrib_df.iloc[[topx - 1]].cumulative.item()
        tot_cum = contrib_df.iloc[[-1]].cumulative.item()
        diff = tot_cum-old_cum

        rest_df = pd.DataFrame(
            {
                'col': ['REST'], 
                'contribution': [diff], 
                'value': ['-'], 
                'cumulative': [tot_cum]
            })

        contrib_df = pd.concat([contrib_df.head(topx), rest_df], axis=0).reset_index(drop=True)

    # add the cumulative before the current variable (i.e. the base of the
    # bar in the graph):
    contrib_df['base']= contrib_df['cumulative'] - contrib_df['contribution']
    return contrib_df


def get_contrib_summary_df(contrib_df, classification=False, round=2):
    """
    returns a DataFrame that summarizes a contrib_df as a pair of
    Reasons+Effect.
    """
    contrib_summary_df = pd.DataFrame(columns=['Reason', 'Effect'])

    for _, row in contrib_df.iterrows():
        if row['col'] == 'base_value':
            reason = 'BASE VALUE'
        else:
            reason = f"{row['col']} = {row['value']}"
        effect = f"{'+' if row['contribution'] >= 0 else ''}"
        if classification:
            effect += str(np.round(100*row['contribution'], round))+'%'
        else:
            effect +=  str(np.round(row['contribution'], round))

        contrib_summary_df = contrib_summary_df.append(
            pd.DataFrame({
                'Reason': [reason],
                'Effect': [effect]
            }))
    return contrib_summary_df.reset_index(drop=True)





def normalize_shap_interaction_values(shap_interaction_values, shap_values=None):
    """
    Normalizes shap_interaction_values to make sure that the rows add up to
    the shap_values.

    This is a workaround for an apparant bug where the diagonals of
    shap_interaction_values of a RandomForestClassifier are set equal to the
    shap_values instead of the main effect.

    Opened an issue here: https://github.com/slundberg/shap/issues/723

    (so far doesn't seem to be fixed)
    """
    siv = shap_interaction_values.copy()

    orig_diags = np.einsum('ijj->ij', siv)
    row_sums = np.einsum('ijk->ij', siv)
    row_diffs = row_sums - orig_diags # sum of rows excluding diagonal elements

    if shap_values is not None:
        diags = shap_values - row_diffs
    else:
        # if no shap_values provided assume that the original diagonal values
        # were indeed equal to the shap values, and so simply
        diags = orig_diags - row_diffs

    s0, s1, s2 = siv.shape

    # should have commented this bit of code earlier:
    #   (can't really explain it anymore, but it works!)
    # In any case, it assigns our news diagonal values to siv:
    siv.reshape(s0,-1)[:,::s2+1] = diags
    return siv


def get_decision_trees(rf_model, X, y):
    """
    Returns a list of ShadowDecTree from the dtreeviz package

    """
    assert hasattr(rf_model, 'estimators_'), \
        """The model does not have an estimators_ attribute, so probably not
        actually a sklearn compatible random forest?"""
    decision_trees = [ShadowDecTree(decision_tree,
                                  X,
                                  y,
                                  feature_names=X.columns.tolist(),
                                  class_names = ['Neg', 'Pos'])
                        for decision_tree in rf_model.estimators_]
    return decision_trees


def get_decisiontree_df(decision_tree, observation, pos_label=1):
    _, nodes = decision_tree.predict(observation)

    decisiontree_df = pd.DataFrame(columns=['node_id', 'average', 'feature',
                                     'value', 'split', 'direction',
                                     'left', 'right', 'diff'])
    if decision_tree.isclassifier()[0]:
        def node_pred_proba(node):
            return node.class_counts()[pos_label]/ sum(node.class_counts())
        for node in nodes:
            if not node.isleaf():
                decisiontree_df = decisiontree_df.append({
                    'node_id' : node.id,
                    'average' : node_pred_proba(node),
                    'feature' : node.feature_name(),
                    'value' : observation[node.feature_name()],
                    'split' : node.split(),
                    'direction' : 'left' if observation[node.feature_name()] < node.split() else 'right',
                    'left' : node_pred_proba(node.left),
                    'right' : node_pred_proba(node.right),
                    'diff' : node_pred_proba(node.left) - node_pred_proba(node) \
                                if observation[node.feature_name()] < node.split() \
                                else node_pred_proba(node.right) - node_pred_proba(node)
                }, ignore_index=True)

    else:
        def node_mean(node):
            return decision_tree.tree_model.tree_.value[node.id].item()
        for node in nodes:
            if not node.isleaf():
                decisiontree_df = decisiontree_df.append({
                    'node_id' : node.id,
                    'average' : node_mean(node),
                    'feature' : node.feature_name(),
                    'value' : observation[node.feature_name()],
                    'split' : node.split(),
                    'direction' : 'left' if observation[node.feature_name()] < node.split() else 'right',
                    'left' : node_mean(node.left),
                    'right' : node_mean(node.right),
                    'diff' : node_mean(node.left) - node_mean(node) \
                                if observation[node.feature_name()] < node.split() \
                                else node_mean(node.right) - node_mean(node)
                }, ignore_index=True)
    return decisiontree_df


def decisiontree_df_summary(decisiontree_df, classifier=False, round=2):
    if classifier:
        base_value = np.round(100*decisiontree_df.iloc[[0]]['average'].item(), round)
        prediction = np.round(100*(decisiontree_df.iloc[[-1]]['average'].item() \
                                + decisiontree_df.iloc[[-1]]['diff'].item()), round)
    else:
        base_value = np.round(decisiontree_df.iloc[[0]]['average'].item(), round)
        prediction = np.round(decisiontree_df.iloc[[-1]]['average'].item() \
                                + decisiontree_df.iloc[[-1]]['diff'].item(), round)


    decisiontree_summary_df = pd.DataFrame(columns=['value', 'condition', 'change', 'prediction'])

    for _, row in decisiontree_df.iterrows():
        if classifier:
            decisiontree_summary_df = decisiontree_summary_df.append({
                            'value' : (str(row['feature'])+'='+str(row['value'])).ljust(50),
                            'condition' : str('>=' if row['direction'] == 'right' else '< ') + str(row['split']).ljust(10),
                            'change' : str('+' if row['diff'] >= 0 else '') + str(np.round(100*row['diff'], round)) +'%',
                            'prediction' : str(np.round(100*(row['average']+row['diff']), round)) + '%'
                        }, ignore_index=True)
        else:
            decisiontree_summary_df = decisiontree_summary_df.append({
                            'value' : (str(row['feature'])+'='+str(row['value'])).ljust(50),
                            'condition' : str('>=' if row['direction'] == 'right' else '< ') + str(row['split']).ljust(10),
                            'change' : str('+' if row['diff'] >= 0 else '') + str(np.round(row['diff'], round)),
                            'prediction' : str(np.round((row['average']+row['diff']), round))
                        }, ignore_index=True)

    return base_value, prediction, decisiontree_summary_df


__all__ = ['DecisionTreesTab']

import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_table

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from .dashboard_methods import *


class DecisionTreesTab:
    def __init__(self, explainer, standalone=False, tab_id="decision_trees", title='Decision Trees',
                 round=2, **kwargs):
        self.explainer = explainer
        self.standalone = standalone
        self.tab_id = tab_id
        self.title = title

        self.round = round
        self.kwargs = kwargs

        if self.standalone:
            self.label_selector = TitleAndLabelSelector(explainer, title=title)
        
        
    def layout(self):
        return dbc.Container([
            self.label_selector.layout() if self.standalone else None,
            # need to add dummy to make callbacks on tab change work:
            html.Div(id='tabs') if self.standalone else None, 
            decision_trees_layout(self.explainer, round=self.round)
        ], fluid=True)
        
    def register_callbacks(self, app):
        if self.standalone:
            self.label_selector.register_callbacks(app)
        decision_trees_callbacks(self.explainer, app, round=self.round)


def decision_trees_layout(explainer, round=2, **kwargs):
    return dbc.Container([
        dbc.Row([
            dbc.Col([
                html.H2('Predictions of individual decision trees.'),
                dbc.Form(
                    [
                        dbc.FormGroup(
                            [
                                dbc.Label("Index", className="mr-2"),
                                dbc.Input(id='tree-input-index', 
                                            placeholder="Fill in index here...",
                                            debounce=True),
                            ],
                            className="mr-3",
                        ),
                        dbc.Button("Random Index", color="primary", id='tree-index-button'),
                    ], inline=True),
                dcc.Store(id='tree-index-store'),
                html.H4('(click on a prediction to see decision path)'),
               # dcc.Loading(id="loading-trees-graph", 
                           # children=[
                                dcc.Graph(id='tree-predictions-graph'),
                                #]),  
            ], width={"size": 8, "offset": 2})
        ]), 
        dbc.Row([
            dbc.Col([
                html.Label('Decision path in decision tree:'),
                dcc.Markdown(id='tree-basevalue'),
                dash_table.DataTable(
                    id='tree-predictions-table',
                    style_cell={'fontSize':20, 'font-family':'sans-serif'},
                ),
            ], width={"size": 6, "offset": 3})
        ]), 
        dbc.Row([
            dbc.Col([
                 dcc.Loading(id="loading-tree-svg-graph", 
                            children=[html.Img(id='dtreeviz-svg')])
            ])
            
        ])
    ],  fluid=True)


def decision_trees_callbacks(explainer, app, round=2, **kwargs):
    @app.callback(
        [Output('tree-predictions-table', 'data'),
         Output('tree-predictions-table', 'columns')],
        [Input('tree-predictions-graph', 'clickData')],
         # Input('label-store', 'data')], # this causes issues for some reason, only on this tab??
        [State('tree-index-store', 'data'),
         State('tabs', 'value')])
    def display_tree_click_data(clickdata, index, tab):
        if clickdata is not None and index is not None:
            tree_idx = int(clickdata['points'][0]['text'].split('tree no ')[1].split(':')[0]) if clickdata is not None else 0
            _, _, decisiontree_df = explainer.decisiontree_df_summary(tree_idx, index, round=round)
            columns = [{'id': c, 'name': c} for c in  decisiontree_df.columns.tolist()]
            return (decisiontree_df.to_dict('records'), columns)
        raise PreventUpdate

    @app.callback(
        Output('dtreeviz-svg', 'src'),
        [Input('tree-predictions-graph', 'clickData'),
         #Input('label-store', 'data')#this causes issues for some reason, only on this tab??
         ],
        [State('tree-index-store', 'data'),
         State('tabs', 'value')])
    def display_click_data(clickData, index, tab):
        if clickData is not None and index is not None and explainer.graphviz_available and explainer.is_classifier:
            tree_idx = int(clickData['points'][0]['text'].split('tree no ')[1].split(':')[0]) 
            svg_encoded = explainer.decision_path_encoded(tree_idx, index)
            return svg_encoded
        return ""

    @app.callback(
        Output('tree-predictions-graph', 'figure'),
        [Input('tree-index-store', 'data'),
         Input('label-store', 'data'),
         Input('tree-predictions-graph', 'clickData')],
        [State('tabs', 'value')]
    )
    def update_tree_graph(index, pos_label, clickdata, tab):
        if index is not None:
            highlight_tree = int(clickdata['points'][0]['text'].split('tree no ')[1].split(':')[0]) if clickdata is not None else None
            return explainer.plot_trees(index, highlight_tree=highlight_tree, round=round)
        return {}

    @app.callback(
        Output('tree-index-store', 'data'),
        [Input('tree-input-index', 'value')],
        [State('tabs', 'value')]
    )
    def update_tree_index_store(index, tab):
        if (explainer.idxs is None 
            and str(index).isdigit() 
            and int(index) >= 0
            and int(index) <= len(explainer)):
            return int(index)
        if (explainer.idxs is not None and index in explainer.idxs):
            return index
        return None

    @app.callback(
        Output('tree-input-index', 'value'),
        [Input('tree-index-button', 'n_clicks')],
        [State('tabs', 'value')]
    )
    def update_tree_input_index(n_clicks, tab):
        return explainer.random_index(return_str=True)


__all__ = ['ContributionsTab']

import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_table

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

import numpy as np
import pandas as pd

from .dashboard_methods import *

class ContributionsTab:
    def __init__(self, explainer, standalone=False, tab_id="contributions", title='Contributions',
                 n_features=15, round=2, **kwargs):
        self.explainer = explainer
        self.standalone = standalone
        self.tab_id = tab_id
        self.title = title
        
        self.n_features = n_features
        self.round = round
        self.kwargs = kwargs
        if self.standalone:
            self.label_selector = TitleAndLabelSelector(explainer, title=title)
           
    def layout(self):
        return dbc.Container([
            self.label_selector.layout() if self.standalone else None,
            # need to add dummy to make callbacks on tab change work:
            html.Div(id='tabs') if self.standalone else None, 
            contributions_layout(self.explainer,  
                    n_features=self.n_features, round=self.round, **self.kwargs)
        ], fluid=True)
    
    def register_callbacks(self, app):
        if self.standalone:
            self.label_selector.register_callbacks(app)
        contributions_callbacks(self.explainer, app, standalone=self.standalone)


def contributions_layout(explainer, n_features=15, round=2, **kwargs):
    """returns layout for individual contributions tabs
    
    :param explainer: ExplainerBunch to build layout for
    :type explainer: ExplainerBunch
    :type title: str
    :param standalone: when standalone layout, include a a label_store, defaults to False
    :type standalone: bool
    :param hide_selector: if model is a classifier, optionally hide the positive label selector, defaults to False
    :type hide_selector: bool
    :param n_features: Default number of features to display in contributions graph, defaults to 15
    :type n_features: int, optional
    :param round: Precision of floats to display, defaults to 2
    :type round: int, optional
    :rtype: [dbc.Container
    """

    if explainer.is_classifier:
        index_choice_form = dbc.Form([

                dbc.FormGroup([
                    html.Div([
                        dbc.Label('Range to select from (prediction probability or prediction percentile):', 
                                    html_for='prediction-range-slider'),
                        dcc.RangeSlider(
                        id='prediction-range-slider',
                        min=0.0, max=1.0, step=0.01,
                        value=[0.5, 1.0],  allowCross=False,
                        marks={0.0:'0.0', 0.1:'0.1', 0.2:'0.2', 0.3:'0.3', 
                                0.4:'0.4', 0.5:'0.5', 0.6:'0.6', 0.7:'0.7', 
                                0.8:'0.8', 0.9:'0.9', 1.0:'1.0'},
                        tooltip = {'always_visible' : False})
                    ], style={'margin-bottom':25})
                    
                ]),
                dbc.FormGroup([
                    dbc.RadioItems(
                        id='include-labels',
                        options=[
                            {'label': explainer.pos_label_str, 'value': 'pos'},
                            {'label': 'Not ' + explainer.pos_label_str, 'value': 'neg'},
                            {'label': 'Both/either', 'value': 'any'},
                        ],
                        value='any',
                        inline=True),
                    dbc.RadioItems(
                        id='preds-or-ranks',
                        options=[
                            {'label': 'Use predictions', 'value': 'preds'},
                            {'label': 'Use percentiles', 'value': 'ranks'},
                        ],
                        value='preds',
                        inline=True)
                ])        
            ])
    else:
        index_choice_form =  dbc.Form([
                dbc.FormGroup([
                    html.Div([
                        html.Div([
                            dbc.Label('Range of predicted outcomes to select from:', 
                                    html_for='prediction-range-slider'),
                            dcc.RangeSlider(
                                id='prediction-range-slider',
                                min=min(explainer.preds), max=max(explainer.preds), 
                                step=np.float_power(10, -round),
                                value=[min(explainer.preds), max(explainer.preds)], 
                                marks={min(explainer.preds):str(np.round(min(explainer.preds), round)),
                                        max(explainer.preds):str(np.round(max(explainer.preds), round))}, 
                                allowCross=False,
                                tooltip = {'always_visible' : False}
                            )

                        ], style={'margin-bottom':25})
                        
                    ]),  
                ], style={'margin-bottom':25}),
            ])

    return dbc.Container([
    dbc.Row([
        dbc.Col([
            html.H2('Display prediction for:'),
            dbc.Input(id='input-index', 
                        placeholder="Fill in index here...",
                        debounce=True),
            index_choice_form,
            dbc.Button("Random Index", color="primary", id='index-button'),
            dcc.Store(id='index-store'),
        ], md=6),
 
        dbc.Col([
             dcc.Loading(id="loading-model-prediction", 
                         children=[dcc.Markdown(id='model-prediction')]),      
        ], md=6),
        
    ], align="start", justify="between"),

    dbc.Row([
        dbc.Col([
            html.H3('Contributions to prediction'),
            dbc.Label('Number of features to display:', html_for='contributions-size'),
            html.Div([
                dcc.Slider(id='contributions-size', 
                    min = 1, max = len(explainer.columns), 
                    marks = {int(i) : str(int(i)) 
                                for i in np.linspace(
                                        1, len(explainer.columns_cats), 6)},
                    step = 1, value=min(n_features, len(explainer.columns_cats)),
                    tooltip = {'always_visible' : False}
                ),
            ], style={'margin-bottom':25}),
            
            dbc.Label('(click on a bar to display pdp graph)'),
            dcc.Loading(id="loading-contributions-graph", 
                        children=[dcc.Graph(id='contributions-graph')]),
            
            html.Div(id='contributions-size-display', style={'margin-top': 20}),
            html.Div(id='contributions-clickdata'),
        ], md=6),

        dbc.Col([
            html.H3('Partial Dependence Plot'),
            dbc.Label("Plot partial dependence plot (\'what if?\') for column:", html_for='pdp-col'),
            dcc.Dropdown(id='pdp-col', 
                options=[{'label': col, 'value':col} 
                            for col in explainer.mean_abs_shap_df(cats=True)\
                                                        .Feature.tolist()],
                value=explainer.mean_abs_shap_df(cats=True).Feature[0]),
            dcc.Loading(id="loading-pdp-graph", 
                    children=[dcc.Graph(id='pdp-graph')]),
        ], md=6)
    ]),
    dbc.Row([
        dbc.Col([
            html.H3('Contributions to prediction'),
            dbc.Label('(table format)'),
            dash_table.DataTable(
                id='contributions_table',
                style_cell={'fontSize':20, 'font-family':'sans-serif'},
                columns=[{'id': c, 'name': c} 
                            for c in ['Reason', 'Effect']],
                      
            ),    
        ], md=6),
    ]),
    ], fluid=True)


def contributions_callbacks(explainer, app, round=2, **kwargs):

    if explainer.is_classifier:
        @app.callback(
            Output('input-index', 'value'),
            [Input('index-button', 'n_clicks')],
            [State('prediction-range-slider', 'value'),
             State('include-labels', 'value'),
             State('preds-or-ranks', 'value'),
             State('tabs', 'value')]
        )
        def update_input_index(n_clicks, slider_range, include, preds_or_ranks, tab):
            y = None
            if include=='neg': y = 0 
            elif include=='pos': y = 1
            return_str = True if explainer.idxs is not None else False

            if preds_or_ranks == 'preds':
                idx = explainer.random_index(
                    y_values=y, pred_proba_min=slider_range[0], pred_proba_max=slider_range[1],
                    return_str=return_str)
            elif preds_or_ranks == 'ranks':
                idx = explainer.random_index(
                    y_values=y, pred_percentile_min=slider_range[0], pred_percentile_max=slider_range[1],
                    return_str=return_str)

            if idx is not None:
                return idx
            raise PreventUpdate
    else:
        @app.callback(
            Output('input-index', 'value'),
            [Input('index-button', 'n_clicks')],
            [State('prediction-range-slider', 'value')]
        )
        def update_input_index(n_clicks, slider_range,):
            y = None
            return_str = True if explainer.idxs is not None else False
            idx = explainer.random_index(pred_min=slider_range[0], pred_max=slider_range[1], return_str=return_str)
            if idx is not None:
                return idx
            raise PreventUpdate

    @app.callback(
        Output('index-store', 'data'),
        [Input('input-index', 'value')]
    )
    def update_bsn_div(input_index):
        if (explainer.idxs is None 
            and str(input_index).isdigit() 
            and int(input_index) <= len(explainer)):
            return int(input_index)
        elif (explainer.idxs is not None
             and str(input_index) in explainer.idxs):
             return str(input_index)
        raise PreventUpdate

    @app.callback(
        [Output('model-prediction', 'children'),
         Output('contributions-graph', 'figure'),
         Output('contributions_table', 'data'),
         Output('contributions_table', 'tooltip_data')],
        [Input('index-store', 'data'),
         Input('contributions-size', 'value'),
         Input('label-store', 'data')]
    )
    def update_output_div(index, topx, pos_label):
        if index is None:
            raise PreventUpdate
        prediction_result_md = explainer.prediction_result_markdown(index)
        plot = explainer.plot_shap_contributions(index, topx=topx, round=round)
        summary_table = explainer.contrib_summary_df(index, round=round)
        tooltip_data = [{'Reason': desc} for desc in explainer.description_list(explainer.contrib_df(index)['col'])]
        return (prediction_result_md, plot, summary_table.to_dict('records'), tooltip_data)

    @app.callback(
        Output('pdp-col', 'value'),
        [Input('contributions-graph', 'clickData')])
    def update_pdp_col(clickData):
        if clickData is not None:
            col = clickData['points'][0]['x']
            return col
        raise PreventUpdate

    @app.callback(
        Output('pdp-graph', 'figure'),
        [Input('index-store', 'data'),
         Input('pdp-col', 'value'),
         Input('label-store', 'data')]
    )
    def update_pdp_graph(idx, col, pos_label):
        explainer.pos_label = pos_label #needed in case of multiple workers
        return explainer.plot_pdp(col, idx, sample=100)


__all__ = ['titanic_survive', 'titanic_fare', 'titanic_names']

import pandas as pd
from pathlib import Path


d_train = pd.read_csv(Path(__file__).resolve().parent / 'datasets'/ 'titanic_train.csv')
d_test = pd.read_csv(Path(__file__).resolve().parent / 'datasets'/'titanic_test.csv')

def titanic_survive():
    X_train = d_train.drop(['Survived', 'Name'], axis=1)
    y_train = d_train['Survived']
    X_test = d_test.drop(['Survived', 'Name'], axis=1)
    y_test = d_test['Survived']
    return X_train, y_train, X_test, y_test

def titanic_fare():
    X_train = d_train.drop(['Fare', 'Name'], axis=1)
    y_train = d_train['Fare']
    X_test = d_test.drop(['Fare', 'Name'], axis=1)
    y_test = d_test['Fare']
    return X_train, y_train, X_test, y_test

def titanic_names():
    return (d_train['Name'].values.tolist(), d_test['Name'].values.tolist())



# -*- coding: utf-8 -*-

__all__ = ['ExplainerDashboard', 'ExplainerDashboardStandaloneTab',
            'ModelSummaryTab']


import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

import plotly.io as pio

from .dashboard_tabs.dashboard_methods import *
from .dashboard_tabs.model_summary_tab import *
from .dashboard_tabs.contributions_tab import *
from .dashboard_tabs.shap_dependence_tab import *
from .dashboard_tabs.shap_interactions_tab import *
from .dashboard_tabs.decision_trees_tab import *


class ExplainerDashboard:
    """Constructs a dashboard out of an ExplainerBunch object. You can indicate
    which tabs to include, and pass kwargs to individual tabs.
    """
    def __init__(self, explainer, title='Model Explainer',   
                tabs=None,
                model_summary=False,  
                contributions=False,
                shap_dependence=False,
                shap_interaction=False,
                decision_trees=False,
                plotly_template="none",
                **kwargs):
        """Constructs an ExplainerDashboard.
        
        :param explainer: an ExplainerBunch object
        :param title: Title of the dashboard, defaults to 'Model Explainer'
        :type title: str, optional
        :param model_summary: display model_summary tab or not, defaults to True
        :type model_summary: bool, optional
        :param contributions: display individual contributions tab or not, defaults to True
        :type contributions: bool, optional
        :param shap_dependence: display shap dependence tab or not, defaults to True
        :type shap_dependence: bool, optional
        :param shap_interaction: display tab interaction tab or not. 
        :type shap_interaction: bool, optional
        :param decision_trees: display tab with individual decision tree of random forest, defaults to False
        :type decision_trees: bool, optional
        """
        self.explainer=explainer
        self.title = title
        self.model_summary = model_summary
        self.contributions = contributions
        self.shap_dependence = shap_dependence
        self.shap_interaction = shap_interaction
        self.decision_trees = decision_trees
        self.plotly_template = plotly_template
        self.kwargs = kwargs

        # calculate lazily loaded properties before starting dashboard:
        if shap_dependence or contributions or model_summary:
            _ = explainer.shap_values, explainer.preds, explainer.pred_percentiles
            if explainer.cats is not None:
                _ = explainer.shap_values_cats
            if explainer.is_classifier:
                _ = explainer.pred_probas
        if model_summary:
            _ = explainer.permutation_importances
            if explainer.cats is not None:
                _ = explainer.permutation_importances_cats
        if shap_interaction:
            _ = explainer.shap_interaction_values
            if explainer.cats is not None:
                _ = explainer.shap_interaction_values_cats
        if decision_trees:
            _ = explainer.graphviz_available
            _ = explainer.decision_trees
            
        self.app = dash.Dash(__name__)
        self.app.config['suppress_callback_exceptions']=True
        self.app.css.config.serve_locally = True
        self.app.scripts.config.serve_locally = True
        self.app.title = title
        
        pio.templates.default = self.plotly_template

        # layout
        self.title_and_label_selector = TitleAndLabelSelector(explainer, title=title)
        self.tabs = [] if tabs is None else tabs
        self._insert_tabs()
        assert len(self.tabs) > 0, 'need to pass at least one tab! e.g. model_summary=True'
        
        self.tab_layouts = [
            dcc.Tab(children=tab.layout(), label=tab.title, id=tab.tab_id, value=tab.tab_id) 
                for tab in self.tabs]

        self.app.layout = dbc.Container([
            self.title_and_label_selector.layout(),
            dcc.Tabs(id="tabs", value=self.tabs[0].tab_id, children=self.tab_layouts),
        ], fluid=True)

        #register callbacks
        self.title_and_label_selector.register_callbacks(self.app)
        
        for tab in self.tabs:
            tab.register_callbacks(self.app)

    def _insert_tabs(self):
        if self.model_summary:
            self.tabs.append(ModelSummaryTab(self.explainer, **self.kwargs))
        if self.contributions:
            self.tabs.append(ContributionsTab(self.explainer, **self.kwargs))
        if self.shap_dependence:
            self.tabs.append(ShapDependenceTab(self.explainer, **self.kwargs))
        if self.shap_interaction:
            self.tabs.append(ShapInteractionsTab(self.explainer, **self.kwargs))
        if self.decision_trees:
            assert hasattr(self.explainer, 'decision_trees'), \
                """the explainer object has no shadow_trees property. This tab 
                only works with a RandomForestClassifierBunch or RandomForestRegressionBunch""" 
            self.tabs.append(DecisionTreesTab(self.explainer, **self.kwargs))
        
    def run(self, port=8050, **kwargs):
        """Starts the dashboard using the built-in Flask server on localhost:port
        
        :param port: the port to run the dashboard on, defaults to 8050
        :type port: int, optional
        """
        print(f"Running {self.title} on http://localhost:{port}")
        pio.templates.default = self.plotly_template
        self.app.run_server(port=port, **kwargs)


class ExplainerDashboardStandaloneTab:
    """Constructs a dashboard out of an ExplainerBunch object. You can indicate
    which tabs to include, and pass kwargs to individual tabs.
    """
    def __init__(self, explainer, tab, title='Model Explainer', 
                    plotly_template="none", **kwargs):
        """Constructs an ExplainerDashboard.
        
        :param explainer: an ExplainerBunch object
        :param title: Title of the dashboard, defaults to 'Model Explainer'
        :type title: str, optional
        :param tab: single tab to be run as dashboard
        """
        self.explainer = explainer
        self.title = title

        self.plotly_template = plotly_template
        self.kwargs = kwargs

        self.tab = tab(self.explainer, standalone=True, **self.kwargs)

        self.app = dash.Dash(__name__)
        self.app.config['suppress_callback_exceptions']=True
        self.app.css.config.serve_locally = True
        self.app.scripts.config.serve_locally = True
        self.app.title = title
        
        pio.templates.default = self.plotly_template

        self.app.layout = self.tab.layout()
        self.tab.register_callbacks(self.app)

    def run(self, port=8050, **kwargs):
        """Starts the dashboard using the built-in Flask server on localhost:port
        
        :param port: the port to run the dashboard on, defaults to 8050
        :type port: int, optional
        """
        print(f"Running {self.title} on http://localhost:{port}")
        pio.templates.default = self.plotly_template
        self.app.run_server(port=port, **kwargs)





__all__ = ['ModelSummaryTab']
import numpy as np

import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from .dashboard_methods import *

from sklearn.metrics import mean_squared_error, mean_absolute_error, r2_score


class ModelSummaryTab:
    def __init__(self, explainer, standalone=False, tab_id="model_summary", title='Model Summary',
                 bin_size=0.1, quantiles=10, cutoff=0.5, 
                 round=2, logs=False, vs_actual=False, ratio=False,
                 n_features=15, **kwargs):
        self.explainer = explainer
        self.standalone = standalone
        
        self.tab_id = tab_id
        self.title = title

        if self.standalone:
            self.label_selector = TitleAndLabelSelector(explainer, title=title)

        if self.explainer.is_classifier:
            self.model_stats = ClassifierModelStats(explainer, bin_size, quantiles, cutoff) 
        elif explainer.is_regression:
            self.model_stats = RegressionModelStats(explainer, round, logs, vs_actual, ratio)
        else:
            self.model_stats =  EmptyLayout()

        self.importances = ImportancesStats(explainer, n_features)

    def layout(self):
        return dbc.Container([
            self.label_selector.layout() if self.standalone else None,
            # need to add dummy to make callbacks on tab change work:
            html.Div(id='tabs') if self.standalone else None, 
            dcc.Store(id='testtest'),
            self.model_stats.layout(),
            self.importances.layout()
        ], fluid=True)

    
    def register_callbacks(self, app, **kwargs):
        if self.standalone:
            self.label_selector.register_callbacks(app)
        self.model_stats.register_callbacks(app)
        self.importances.register_callbacks(app)


class ImportancesStats:
    def __init__(self, explainer, n_features=None):
        self.explainer = explainer
        self.n_features = n_features
    
    def layout(self):
        cats_display = 'none' if self.explainer.cats is None else None
        return dbc.Container([
            dbc.Row([dbc.Col([html.H2('Feature Importances:')])]),
            dbc.Row([
                dbc.Col([
                    dbc.FormGroup(
                            [
                                dbc.Label("Importances type:"),
                                dbc.RadioItems(
                                    options=[
                                        {'label': 'Permutation Importances', 
                                        'value': 'permutation'},
                                        {'label': 'SHAP values', 
                                        'value': 'shap'}
                                    ],
                                    value='shap',
                                    id='permutation-or-shap',
                                    inline=True,
                                ),
                            ]
                        )
                ]),
                dbc.Col([
                    html.Label('Select max number of importances to display:'),
                    dcc.Dropdown(id='importance-tablesize',
                                        options = [{'label': str(i+1), 'value':i+1} 
                                                    for i in range(len(self.explainer.columns_cats))],
                                        value=min(self.n_features, len(self.explainer.columns_cats)))
                ]),
                dbc.Col([
                    html.Div([
                        dbc.Label("Grouping:"),
                        dbc.FormGroup(
                        [
                            dbc.RadioButton(
                                id='group-categoricals', 
                                className="form-check-input",
                                checked=True),
                            dbc.Label("Group Cats",
                                    html_for='group-categoricals',
                                    className="form-check-label"),
                        ], check=True),
                    ], style=cats_display)
                ]),
                    
            ], form=True, justify="between"),

            dbc.Row([
                dbc.Col([
                    dcc.Loading(id="loading-importances-graph", 
                            children=[dcc.Graph(id='importances-graph')])
                ]),
            ]), 
            ], fluid=True)
        
    def register_callbacks(self, app, **kwargs):
        @app.callback(  
            Output('importances-graph', 'figure'),
            [Input('importance-tablesize', 'value'),
             Input('group-categoricals', 'checked'),
             Input('permutation-or-shap', 'value'),
             Input('label-store', 'data')],
            [State('tabs', 'value')]
        )
        def update_importances(tablesize, cats, permutation_shap, pos_label, tab): 
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            return self.explainer.plot_importances(
                        kind=permutation_shap, topx=tablesize, cats=cats)

class ClassifierModelStats:
    def __init__(self, explainer, bin_size=0.1, quantiles=10, cutoff=0.5):
        self.explainer = explainer
        self.bin_size, self.quantiles, self.cutoff = bin_size, quantiles, cutoff

    def layout(self):
        return dbc.Container([
            dbc.Row([dbc.Col([html.H2('Model Performance:')])]),

            dbc.Row([
                dbc.Col([
                    html.Div([
                        dcc.Loading(id="loading-precision-graph", 
                                children=[dcc.Graph(id='precision-graph')]),
                    ], style={'margin': 0}),
                    html.Div([
                        dbc.Label('Bin size:', html_for='precision-binsize'),
                        html.Div([
                            dcc.Slider(id='precision-binsize', 
                                    min = 0.01, max = 0.5, step=0.01, value=self.bin_size,
                                    marks={0.01: '0.01', 0.05: '0.05', 0.10: '0.10',
                                        0.20: '0.20', 0.25: '0.25' , 0.33: '0.33', 
                                        0.5: '0.5'}, 
                                    included=False,
                                    tooltip = {'always_visible' : False})
                        ], style={'margin-bottom': 25}),
                    ], id='bin-size-div', style={'margin': 5}),
                    html.Div([
                        dbc.Label('Quantiles:', html_for='precision-quantiles'),
                        html.Div([
                            dcc.Slider(id='precision-quantiles', 
                                        min = 1, max = 20, step=1, value=self.quantiles,
                                        marks={1: '1', 5: '5', 10: '10', 15: '15', 20:'20'}, 
                                        included=False,
                                        tooltip = {'always_visible' : False}),
                        ], style={'margin-bottom':25}),
                    ], id='quantiles-div', style={'margin': 5}),
                    dbc.Label('Binning Method:', html_for='binsize-or-quantiles'),
                    dbc.RadioItems(
                        id='binsize-or-quantiles',
                        options=[
                            {'label': 'Bin Size', 
                            'value': 'bin_size'},
                            {'label': 'Quantiles', 
                            'value': 'quantiles'}
                        ],
                        value='bin_size',
                        inline=True),
                    dbc.FormGroup([
                                dbc.RadioButton(
                                    id="precision-multiclass", className="form-check-input"
                                ),
                                dbc.Label(
                                    "Display all classes",
                                    html_for="precision-multiclass",
                                    className="form-check-label",
                                ),
                            ], check=True),
                ], md=6, align="start"),
                dbc.Col([
                    dcc.Loading(id="loading-confusionmatrix-graph", 
                                children=[dcc.Graph(id='confusionmatrix-graph')]),
                    dbc.FormGroup([
                                dbc.RadioButton(
                                    id='confusionmatrix-percentage', 
                                    className="form-check-input", 
                                    checked=True
                                ),
                                dbc.Label(
                                    "Display percentages",
                                    html_for="confusionmatrix-percentage",
                                    className="form-check-label",
                                ),
                    ], check=True),
                    dbc.FormGroup([
                                dbc.RadioButton(
                                    id="confusionmatrix-binary", 
                                    className="form-check-input", 
                                    checked=True
                                ),
                                dbc.Label(
                                    "Binary (use cutoff for positive vs not positive)",
                                    html_for="confusionmatrix-binary",
                                    className="form-check-label",
                                ),
                    ], check=True),
                ], md=6, align="start"),              
            ]),
            dbc.Row([
                    dbc.Col([
                        html.Div([
                            html.Label('Cutoff prediction probability:'),
                            dcc.Slider(id='precision-cutoff', 
                                        min = 0.01, max = 0.99, step=0.01, value=self.cutoff,
                                        marks={0.01: '0.01', 0.25: '0.25', 0.50: '0.50',
                                                0.75: '0.75', 0.99: '0.99'}, 
                                        included=False,
                                        tooltip = {'always_visible' : False})
                        ], style={'margin-bottom': 25}),
                    ])
                ]),
            dbc.Row([
                    dbc.Col([
                        html.Div([
                            html.Label('Cutoff percentile of samples:'),
                            dcc.Slider(id='percentile-cutoff', 
                                        min = 0.01, max = 0.99, step=0.01, value=self.cutoff,
                                        marks={0.01: '0.01', 0.25: '0.25', 0.50: '0.50',
                                                0.75: '0.75', 0.99: '0.99'}, 
                                        included=False,
                                        tooltip = {'always_visible' : False})
                        ], style={'margin-bottom': 25}),
                    ])
                ]),
            dbc.Row([
                dbc.Col([
                    html.Div([
                        dcc.Loading(id="loading-lift-curve", 
                                children=[dcc.Graph(id='lift-curve-graph')]),
                    ], style={'margin': 0}),
                    dbc.FormGroup([
                        dbc.RadioButton(
                            id="lift-curve-percentage", 
                            className="form-check-input", 
                            checked=True
                        ),
                        dbc.Label(
                            "Display percentages",
                            html_for="lift-curve-percentage",
                            className="form-check-label",
                        ),
                    ], check=True),          
                ], md=6, align="start"),
                dbc.Col([
                    html.Div([
                                dcc.Loading(id="loading-classification-graph", 
                                            children=[dcc.Graph(id='classification-graph')]),
                    ], style={'margin': 0}),

                    dbc.FormGroup([
                                dbc.RadioButton(
                                    id="classification-percentage", 
                                    className="form-check-input", 
                                    checked=True
                                ),
                                dbc.Label(
                                    "Display percentages",
                                    html_for="classification-percentage",
                                    className="form-check-label",
                                ),
                    ], check=True),
                ], md=6, align="start"),
            ]),
            dbc.Row([    
                dbc.Col([
                    dcc.Loading(id="loading-roc-auc-graph", 
                                children=[dcc.Graph(id='roc-auc-graph')]),
                ], md=6),
                dbc.Col([
                    dcc.Loading(id="loading-pr-auc-graph", 
                                children=[dcc.Graph(id='pr-auc-graph')]),
                ], md=6),
            ]),
        ], fluid=True)

    def register_callbacks(self, app, **kwargs):

        @app.callback(
            [Output('bin-size-div', 'style'),
             Output('quantiles-div', 'style')],
            [Input('binsize-or-quantiles', 'value')],
        )
        def update_div_visibility(bins_or_quantiles):
            if bins_or_quantiles=='bin_size':
                return {}, {'display': 'none'}
            elif bins_or_quantiles=='quantiles':
                return {'display': 'none'}, {}
            raise PreventUpdate   
            
        @app.callback(
            Output('lift-curve-graph', 'figure'),
            [Input('lift-curve-percentage', 'checked'),
             Input('precision-cutoff', 'value'),
             Input('label-store', 'data')],
            [State('tabs', 'value')],
        )
        def update_precision_graph(percentage, cutoff, pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            return self.explainer.plot_lift_curve(cutoff=cutoff, percentage=percentage)

        @app.callback(
            Output('precision-graph', 'figure'),
            [Input('precision-binsize', 'value'),
             Input('precision-quantiles', 'value'),
             Input('binsize-or-quantiles', 'value'),
             Input('precision-cutoff', 'value'),
             Input('precision-multiclass', 'checked'),
             Input('label-store', 'data')],
            [State('tabs', 'value')],
        )
        def update_precision_graph(bin_size, quantiles, bins, cutoff, multiclass, pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            if bins=='bin_size':
                return self.explainer.plot_precision(
                    bin_size=bin_size, cutoff=cutoff, multiclass=multiclass)
            elif bins=='quantiles':
                return self.explainer.plot_precision(
                    quantiles=quantiles, cutoff=cutoff, multiclass=multiclass)
            raise PreventUpdate

        @app.callback(
            Output('classification-graph', 'figure'),
            [Input('classification-percentage', 'checked'),
             Input('precision-cutoff', 'value'),
             Input('label-store', 'data')],
            [State('tabs', 'value')],
        )
        def update_precision_graph(percentage, cutoff, pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            return self.explainer.plot_classification(cutoff=cutoff, percentage=percentage)

        @app.callback(
             Output('confusionmatrix-graph', 'figure'),
            [Input('precision-cutoff', 'value'),
             Input('confusionmatrix-percentage', 'checked'),
             Input('confusionmatrix-binary', 'checked'),
             Input('label-store', 'data')],
            [State('tabs', 'value')],
        )
        def update_precision_graph(cutoff, normalized, binary, pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            return self.explainer.plot_confusion_matrix(
                        cutoff=cutoff, normalized=normalized, binary=binary)

        @app.callback(
            Output('roc-auc-graph', 'figure'),
            [Input('precision-cutoff', 'value'),
             Input('label-store', 'data'),
             Input('tabs', 'value')],
        )
        def update_precision_graph(cutoff, pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            return self.explainer.plot_roc_auc(cutoff=cutoff)

        @app.callback(
            Output('pr-auc-graph', 'figure'),
            [Input('precision-cutoff', 'value'),
             Input('label-store', 'data')],
            [State('tabs', 'value')],
        )
        def update_precision_graph(cutoff, pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            return self.explainer.plot_pr_auc(cutoff=cutoff)

        @app.callback(
            Output('precision-cutoff', 'value'),
            [Input('percentile-cutoff', 'value')]
        )
        def update_cutoff(percentile):
            return np.round(self.explainer.cutoff_from_percentile(percentile), 2)

class RegressionModelStats:
    def __init__(self, explainer, round=2, logs=False, vs_actual=False, ratio=False):
        self.explainer = explainer
        self.round, self.logs, self.vs_actual, self. ratio  = round, logs, vs_actual, ratio

    def layout(self):
        return dbc.Container([
            dbc.Row([dbc.Col([html.H2('Model Performance:')])]),
            dbc.Row([
                dbc.Col([
                    
                    dcc.Loading(id="loading-predicted-vs-actual-graph", 
                                children=[dcc.Graph(id='predicted-vs-actual-graph')]),
                    dbc.FormGroup(
                    [
                        dbc.RadioButton(
                            id='preds-vs-actual-logs',
                            className="form-check-input"),
                        dbc.Label("Take Logs",
                                html_for='preds-vs-actual-logs',
                                className="form-check-label"),
                    ], check=True),
                    ], width=6),
                dbc.Col([
                    dcc.Loading(id="loading-model-summary", 
                                children=[dcc.Markdown(id='model-summary')]),      
                ], width=6),
            ], align="start"),
            dbc.Row([
                dbc.Col([

                    dcc.Loading(id="loading-residuals-graph", 
                                children=[dcc.Graph(id='residuals-graph')]),
                    dbc.FormGroup(
                    [
                        dbc.RadioItems(
                            options=[
                                {"label": "vs Prediction", "value": "vs_pred"},
                                {"label": "vs Actual", "value": "vs_actual"},
                            ],
                            value="vs_pred",
                            id='residuals-pred-or-actual',
                            inline=True,
                        ),
                    ]),
                    dbc.FormGroup(
                    [
                        dbc.RadioButton(
                            id='residuals-ratio',
                            className="form-check-input"),
                        dbc.Label("Display Ratio",
                                html_for='residuals-ratio',
                                className="form-check-label"),
                    ], check=True),

                ], width=6),
                dbc.Col([
                    dcc.Loading(id="loading-residuals-vs-col-graph", 
                                children=[dcc.Graph(id='residuals-vs-col-graph')]),
                    dbc.Label("Column:"),
                    dcc.Dropdown(id='residuals-col',
                        options = [{'label': col, 'value': col} 
                                        for col in self.explainer.mean_abs_shap_df(cats=False)\
                                                        .Feature.tolist()],
                        value=self.explainer.mean_abs_shap_df(cats=False)\
                                                        .Feature.tolist()[0]),
                    dbc.FormGroup(
                    [
                        dbc.RadioButton(
                            id='residuals-vs-col-ratio',
                            className="form-check-input"),
                        dbc.Label("Display Ratio",
                                html_for='residuals-vs-col-ratio',
                                className="form-check-label"),
                    ], check=True),
                ], width=6),
            ])
        ], fluid=True)
        

    def register_callbacks(self, app, **kwargs):
        @app.callback(
            Output('model-summary', 'children'),
            [Input('label-store', 'data')],
            [State('tabs', 'value')]
        )
        def update_model_summary(pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            return self.explainer.metrics_markdown()

        Output('model-prediction', 'children')
        @app.callback(
            Output('predicted-vs-actual-graph', 'figure'),
            [Input('preds-vs-actual-logs', 'checked'),
             Input('label-store', 'data')],
            [State('tabs', 'value')]
        )
        def update_predicted_vs_actual_graph(logs, pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            return self.explainer.plot_predicted_vs_actual(logs=logs)

        @app.callback(
            Output('residuals-graph', 'figure'),
            [Input('residuals-pred-or-actual', 'value'),
             Input('residuals-ratio', 'checked'),
             Input('label-store', 'data')],
            [State('tabs', 'value')],
        )
        def update_residuals_graph(pred_or_actual, ratio, pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            vs_actual = pred_or_actual=='vs_actual'
            return self.explainer.plot_residuals(vs_actual=vs_actual, ratio=ratio)

        @app.callback(
            Output('residuals-vs-col-graph', 'figure'),
            [Input('residuals-col', 'value'),
             Input('residuals-vs-col-ratio', 'checked'),
             Input('label-store', 'data')],
            [State('tabs', 'value')],
        )
        def update_residuals_graph(col, ratio, pos_label, tab):
            self.explainer.pos_label = pos_label #needed in case of multiple workers
            return self.explainer.plot_residuals_vs_feature(col, ratio=ratio, dropna=True)






import inspect

import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate


# Stolen from https://www.fast.ai/2019/08/06/delegation/
# then extended to deal with multiple inheritance
def delegates(to=None, keep=False):
    "Decorator: replace `**kwargs` in signature with params from `to`"
    def _f(f):
        from_f = f.__init__ if to is None else f
        sig = inspect.signature(from_f)
        sigd = dict(sig.parameters)
        k = sigd.pop('kwargs')
        if to is None:
            for base_cls in f.__bases__:
                to_f = base_cls.__init__
                s2 = {k:v for k,v in inspect.signature(to_f).parameters.items()
                    if v.default != inspect.Parameter.empty and k not in sigd}
                sigd.update(s2)
        else:
            to_f = to
            s2 = {k:v for k,v in inspect.signature(to_f).parameters.items()
                if v.default != inspect.Parameter.empty and k not in sigd}
            sigd.update(s2)
        if keep: sigd['kwargs'] = k
        from_f.__signature__ = sig.replace(parameters=sigd.values())
        return f
    return _f


class EmptyLayout:
    def __init__(self):
        pass

    def layout(self):
        return None

    def register_callbacks(self, app):
        pass

class TitleAndLabelSelector:
    def __init__(self, explainer, title="Explainer Dashboard", 
                    title_only=False, label_store_only=False, dummy=False):
        self.explainer = explainer
        self.title = title
        self.title_only, self.label_store_only = title_only, label_store_only
        self.dummy = dummy

    def layout(self):
        if self.dummy: return None
    
        title_col = dbc.Col([html.H1(self.title)], width='auto')
    
        if self.title_only:
            return dbc.Container([
                dbc.Row([
                    title_col,
                ], justify="start", align="center")
            ], fluid=True)
        elif self.label_store_only or not self.explainer.is_classifier:
            return dbc.Container([
                dbc.Row([
                    title_col,
                    dcc.Store(id='label-store')
                ], justify="start", align="center")
            ], fluid=True)
        elif self.explainer.is_classifier: 
            #classifier can show label selector, but still optionally hide it
            return dbc.Container([
                dbc.Row([
                    title_col,
                    dbc.Col([
                            dcc.Dropdown(
                                id='pos-label-selector',
                                options = [{'label': label, 'value': label} 
                                                for label in self.explainer.labels],
                                value = self.explainer.pos_label_str
                            )
                    ], width=2),
                    dcc.Store(id='label-store')
                ], justify="start", align="center")
            ], fluid=True)
        else:
            return None
 

    def register_callbacks(self, app, **kwargs):
        if self.explainer.is_classifier and not self.label_store_only:
            @app.callback(
                Output("label-store", "data"),
                [Input('pos-label-selector', 'value')]
            )
            def change_positive_label(pos_label):
                if pos_label is not None:
                    self.explainer.pos_label = pos_label
                    return pos_label
                return self.explainer.pos_label



__all__ = ['ShapInteractionsTab']

import dash
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
import dash_table

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from .dashboard_methods import *


class ShapInteractionsTab:
    def __init__(self, explainer, standalone=False, tab_id="shap_interactions", title='Shap Interactions',
                 n_features=10, cats=True, **kwargs):
        self.explainer = explainer
        self.standalone = standalone
        self.tab_id = tab_id
        self.title = title

        self.n_features = n_features
        self.cats = cats

        self.kwargs = kwargs
        if self.standalone:
            self.label_selector = TitleAndLabelSelector(explainer, title=title)
        
    def layout(self):
        return dbc.Container([
            self.label_selector.layout() if self.standalone else None,
            # need to add dummy to make callbacks on tab change work:
            html.Div(id='tabs') if self.standalone else None, 
            shap_interactions_layout(self.explainer, n_features=self.n_features, cats=self.cats, **self.kwargs)
        ], fluid=True)
    
    def register_callbacks(self, app):
        if self.standalone:
            self.label_selector.register_callbacks(app)
        shap_interactions_callbacks(self.explainer, app)


def shap_interactions_layout(explainer, 
            n_features=10, cats=True, **kwargs):
    """return layout for shap interactions tab.
    
    :param explainer: ExplainerBunch
    :type explainer: ExplainerBunch 
    :param title: title displayed on top of page, defaults to None
    :type title: str
    :param standalone: when standalone layout, include a a label_store, defaults to False
    :type standalone: bool
    :param hide_selector: if model is a classifier, optionally hide the positive label selector, defaults to False
    :type hide_selector: bool
    :param n_features: default number of features to display, defaults to 10
    :type n_features: int, optional
    :rtype: dbc.Container
    """
    cats_display = 'none' if explainer.cats is None else 'inline-block'
    return dbc.Container([
    #dbc.Row([dbc.Col([html.H3('Shap Interaction Values')])]),
    dbc.Row([
        dbc.Col([
            html.H3('Shap Interaction Summary'),
            dbc.Row([
                dbc.Col([
                    dbc.Label("Feature"),
                    dcc.Dropdown(id='interaction-col', 
                        options=[{'label': col, 'value': col} 
                                    for col in explainer.columns_ranked_by_shap(cats)],
                        value=explainer.columns_ranked_by_shap(cats)[0])],
                    md=4), 
                dbc.Col([
                    dbc.Label("Depth:"),
                    dcc.Dropdown(id='interaction-summary-depth',
                        options = [{'label': str(i+1), 'value':i+1} 
                                        for i in range(len(explainer.columns_ranked_by_shap(cats))-1)],
                        value=min(n_features, len(explainer.columns_ranked_by_shap(cats))-1))],
                    md=2), 
                dbc.Col([
                    dbc.FormGroup(
                        [
                            dbc.Label("Summary Type"),
                            dbc.RadioItems(
                                options=[
                                    {"label": "Aggregate", "value": "aggregate"},
                                    {"label": "Detailed", "value": "detailed"},
                                ],
                                value="aggregate",
                                id="interaction-summary-type",
                                inline=True,
                            ),
                        ]
                    )
                ], md=3),
                dbc.Col([
                    dbc.Label("Grouping:"),
                    dbc.FormGroup(
                    [
                        dbc.RadioButton(
                            id='interaction-group-categoricals', 
                            className="form-check-input",
                            checked=cats),
                        dbc.Label("Group Cats",
                                html_for='interaction-group-categoricals',
                                className="form-check-label"),
                    ], check=True)],
                    md=3),
                ], form=True),
            dbc.Label('(Click on a dot to display interaction graph)'),
            dcc.Loading(id="loading-interaction-summary-scatter", 
                         children=[dcc.Graph(id='interaction-shap-summary-graph')])
        ], md=6),
        dbc.Col([
            html.H3('Shap Interaction Plots'),
            dbc.Row([
                dbc.Col([
                    dbc.Label("Interaction Feature"),
                    dcc.Dropdown(id='interaction-interact-col', 
                        options=[{'label': col, 'value':col} 
                                    for col in explainer.columns_ranked_by_shap(cats)],
                        value=explainer.shap_top_interactions(explainer.columns_ranked_by_shap(cats)[0], cats=cats)[1]
                    ),
                ], md=8), 
                dbc.Col([
                    dbc.Label("Highlight index:"),
                    dbc.Input(id='interaction-highlight-index', 
                        placeholder="Highlight index...", debounce=True)],
                    md=4), 
                ], form=True),
            
            dcc.Loading(id="loading-interaction-graph", 
                         children=[dcc.Graph(id='interaction-graph')]),
            dcc.Loading(id="loading-reverse-interaction-graph", 
                         children=[dcc.Graph(id='reverse-interaction-graph')]),
        ], md=6)
    ]), 
    ], fluid=True)


def shap_interactions_callbacks(explainer, app, standalone=False, n_features=10, **kwargs):

    @app.callback(
        [Output('interaction-col', 'options'),
         Output('interaction-col', 'value'),
         Output('interaction-summary-depth', 'options'),
         Output('interaction-summary-depth', 'value')],
        [Input('interaction-group-categoricals', 'checked')],
        [State('interaction-col', 'value'),
         State('tabs', 'value')])
    def update_col_options(cats, col, tab):
        cols = explainer.columns_ranked_by_shap(cats)
        col_options = [{'label': col, 'value': col} for col in cols] 
        if col not in cols:
            # if currently selected col is not in new col options (i.e. because
            # grouped cat and switched to onehot or reverse), get the equivalent
            # col
            col = explainer.inverse_cats(col)
        depth_options = [{'label': str(i+1), 'value': i+1} for i in range(len(cols))]
        depth = len(cols)-1
        return (col_options, col, depth_options, depth)

    @app.callback(
        [Output('interaction-shap-summary-graph', 'figure'),
         Output('interaction-interact-col', 'options')],
        [Input('interaction-summary-type', 'value'),
         Input('interaction-col', 'value'),
         Input('interaction-summary-depth', 'value'),
         Input('label-store', 'data')],
        [State('interaction-group-categoricals', 'checked')])
    def update_interaction_scatter_graph(summary_type, col, depth, pos_label, cats):
        if col is not None:
            explainer.pos_label = pos_label #needed in case of multiple workers
            if depth is None: 
                depth = len(explainer.columns_ranked_by_shap(cats))-1 
            if summary_type=='aggregate':
                plot = explainer.plot_interactions(col, topx=depth, cats=cats)
            elif summary_type=='detailed':
                plot = explainer.plot_shap_interaction_summary(col, topx=depth, cats=cats)

            interact_cols = explainer.shap_top_interactions(col, cats=cats)
            interact_col_options = [{'label': col, 'value':col} for col in interact_cols]
            return plot, interact_col_options
        return None, None

    @app.callback(
        [Output('interaction-highlight-index', 'value'),
         Output('interaction-interact-col', 'value')],
        [Input('interaction-shap-summary-graph', 'clickData')])
    def display_scatter_click_data(clickdata):
        if clickdata is not None and clickdata['points'][0] is not None:
            if isinstance(clickdata['points'][0]['y'], float): # detailed
                idx = clickdata['points'][0]['pointIndex']
                col = clickdata['points'][0]['text'].split('=')[0]                             
                return (idx, col)
            elif  isinstance(clickdata['points'][0]['y'], str): # aggregate
                col = clickdata['points'][0]['y'].split(' ')[1]
                return (dash.no_update, col) 
        raise PreventUpdate

    @app.callback(
        [Output('interaction-graph', 'figure'),
         Output('reverse-interaction-graph', 'figure')],
        [Input('interaction-interact-col', 'value'),
         Input('interaction-highlight-index', 'value'),
         Input('label-store', 'data')],
        [State('interaction-col', 'value'),
         State('interaction-group-categoricals', 'checked')])
    def update_dependence_graph(interact_col, index, pos_label, col, cats):
        if interact_col is not None:
            explainer.pos_label = pos_label #needed in case of multiple workers
            return (explainer.plot_shap_interaction(
                        col, interact_col, highlight_idx=index, cats=cats),
                    explainer.plot_shap_interaction(
                        interact_col, col, highlight_idx=index, cats=cats))
        raise PreventUpdate